# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Base Algorithm for lattice Boltzmann methods
============================================

This module describes various methods involved during
the computation of a solution using lattice Boltzmann
methods.

The default kernels defined here are:

    - transport kernel
    - f2m kernel
    - m2f_kernel
    - equilibrium kernel
    - relaxation kernel
    - source terms kernel
    - one_time_step kernel which makes the stream + source terms + relaxation

You can modify each functions to define your own behavior. This is what is done
when we want to specialize our one_time_step kernel for Pull algorithm,
Push algorithm, ... All these kernels are defined using SymPy and thus must be
expressed as symbolic expressions. Then, we will generate the numerical code.

These kernels are defined in a class (the base class is BaseAlgorithm) where we
have all the needed information of our scheme.

Let's take an example: the f2m kernel allows to compute the moments from the
distribution functions thanks to the matrix M define in our scheme.

The formula is straightforward

    m = M f

where m is the moments, f the distributed functions and M the matrix build with
the polynoms defining the moments of our scheme.

The SymPy expression can be written as

    Eq(m, M*f)

where m and f are symbol matrices and M a SymPy matrix.

When we define a kernel, we make that in two steps. First we define locally
the expression (what we have to do for each point). Then we define the kernel
using this local kernel for the whole domain (we write the loop).

Therefore, for our example we have

    def f2m_local(self, f, m):
        return Eq(m, self.M*f)

    def f2m(self):
        space_index = self._get_space_idx_full()
        f = self._get_indexed_1('f', space_index)
        m = self._get_indexed_1('m', space_index)
        return {'code': For(space_index, self.f2m_local(f, m))}

"""

import sympy as sp
from sympy import Eq

from ..generator import For, If
from ..symbolic import ix, iy, iz, nx, ny, nz, nv, indexed, space_idx, alltogether, recursive_sub
from ..symbolic import rel_ux, rel_uy, rel_uz
from .transform import parse_expr
from .ode import euler
from ..monitoring import monitor


class BaseAlgorithm:
    def __init__(self, scheme, sorder, generator, settings=None):
        xx, yy, zz = sp.symbols('xx, yy, zz')
        self.symb_coord_local = [xx, yy, zz]
        self.symb_coord = scheme.symb_coord
        self.dim = scheme.dim
        self.ns = scheme.stencil.nv_ptr[-1]
        self.M = scheme.M.subs(scheme.param.items())
        self.invM = scheme.invM.subs(scheme.param.items())
        self.all_velocities = scheme.stencil.get_all_velocities()
        self.mv = sp.MatrixSymbol('m', self.ns, 1)

        if scheme.rel_vel is not None:
            self.rel_vel_symb = [rel_ux, rel_uy, rel_uz][:self.dim]
            self.rel_vel = sp.Matrix(scheme.rel_vel)

            self.Tu = scheme.Tu.subs(scheme.param.items())
            self.Tmu = scheme.Tmu.subs(scheme.param.items())

            self.Mu = self.Tu * self.M
            self.invMu = self.invM * self.Tmu
            alltogether(self.Mu, nsimplify=True)
            alltogether(self.invMu, nsimplify=True)
        else:
            self.rel_vel_symb = None

        self.consm = {}
        for k, v in scheme.consm.items():
            self.consm[str(k)] = v

        self.sorder = sorder
        self.generator = generator

        subs_coords = list(zip(self.symb_coord, self.symb_coord_local))
        subs_moments = list(zip(scheme.consm.keys(), [self.mv[int(i), 0] for i in scheme.consm.values()]))
        to_subs = subs_coords + list(scheme.param.items())
        to_subs_full = to_subs + subs_moments

        self.eq = recursive_sub(scheme.EQ, to_subs_full)
        self.s = recursive_sub(scheme.s, to_subs_full)
        alltogether(self.eq, nsimplify=True)
        alltogether(self.s)

        if self.rel_vel_symb:
            self.rel_vel = recursive_sub(self.rel_vel, to_subs_full)
            alltogether(self.rel_vel)

        self.source_eq = []
        for source in scheme._source_terms:
            if source:
                for k, v in source.items():
                    lhs = recursive_sub(k, to_subs_full)
                    if isinstance(v, (float, int)):
                        rhs = v
                    else:
                        rhs = recursive_sub(v, to_subs)
                    self.source_eq.append((lhs, rhs))

        self.vmax = [0]*3
        self.vmax[:scheme.dim] = scheme.stencil.vmax
        self.local_vars = self.symb_coord_local[:self.dim]
        self.settings = settings if settings else {}

    def _get_space_idx_full(self):
        """
        Return a list of SymPy Idx ordered with sorder
        and with the dimensions.

            ix -> [0, nx[
            iy -> [0, ny[
            iz -> [0, nz[

        The length of the list is the dimension of the problem.
        """
        return space_idx([(0, nx), (0, ny), (0, nz)], priority=self.sorder[1:])

    def _get_space_idx_inner(self):
        """
        Return a list of SymPy Idx ordered with sorder
        and with the dimensions.

            ix -> [vmax_x, nx-vmax_x[
            iy -> [vmax_y, ny-vmax_y[
            iz -> [vmax_z, nz-vmax_z[

        where vmax_i is the maximum of the velocities modulus in direction i.
        The length of the list is the dimension of the problem.
        """
        return space_idx([(self.vmax[0], nx-self.vmax[0]),
                          (self.vmax[1], ny-self.vmax[1]),
                          (self.vmax[2], nz-self.vmax[2])],
                         priority=self.sorder[1:])

    def _get_indexed_on_range(self, name, space_index):
        """
        Return a SymPy matrix of indexed objects
        (one component for each velocity index).

        Parameters
        ----------

        name : string
            name of the SymPy symbol for the indexed object

        space_index : list
            list of SymPy Idx corresponding to space variables

        Return
        ------

        SymPy Matrix
            indexed objects for each velocity

        """
        return indexed(name, [self.ns, nx, ny, nz],
                       [nv] + space_index,
                       velocities_index=range(self.ns), priority=self.sorder)

    def _get_indexed_on_velocities(self, name, space_index, velocities):
        """
        Return a SymPy matrix of indexed objects (one component for each velocity).

        Parameters
        ----------

        name : string
            name of the SymPy symbol for the indexed object

        space_index : list
            list of SymPy Idx corresponding to space variables

        velocities : list
            list of velocity components

        Return
        ------

        SymPy Matrix
            indexed objects for each velocity

        """
        return indexed(name, [self.ns, nx, ny, nz],
                       [nv] + space_index,
                       velocities=velocities, priority=self.sorder)

    def relative_velocity(self, m):
        rel_vel = sp.Matrix(self.rel_vel).subs(list(zip(self.mv, m)))
        return [Eq(self.rel_vel_symb[i], rel_vel[i]) for i in range(self.dim)]

    def restore_conserved_moments(self, m, f):
        nconsm = len(self.consm)

        if isinstance(m[nconsm:], list):
            m_consm = sp.Matrix(m[:nconsm])
        else:
            m_consm = m[:nconsm]

        return Eq(m_consm, sp.Matrix((self.Mu*f)[:nconsm]))

    def coords(self):
        coord = []
        for x in self.symb_coord[:self.dim]:
            coord.append(indexed(x.name, [self.ns, nx, ny, nz], priority=self.sorder[1:]))
        return [Eq(xx, x) for xx, x in zip(self.symb_coord_local[:self.dim], coord)]

    def transport_local(self, f, fnew):
        """
        Return the symbolic expression of the lbm transport.

        Parameters
        ----------

        f : SymPy Matrix
            indexed objects of rhs for the distributed functions

        fnew : SymPy Matrix
            indexed objects of lhs for the distributed functions

        """
        return Eq(fnew, f)

    def transport(self):
        """
        Return the code expression of the lbm transport on the whole inner domain.
        """
        space_index = self._get_space_idx_inner()
        f = self._get_indexed_on_velocities('f', space_index, -self.all_velocities)
        fnew = self._get_indexed_on_range('fnew', space_index)
        return {'code': For(space_index, self.transport_local(f, fnew))}

    def f2m_local(self, f, m, with_rel_velocity=False):
        """
        Return symbolic expression which computes the moments from the
        distributed functions.

        Parameters
        ----------

        f : SymPy Matrix
            indexed objects for the distributed functions

        m : SymPy Matrix
            indexed objects for the moments

        with_rel_velocity : boolean
            check if the scheme uses relative velocity.
            If yes, the conserved moments must be computed first.
            (default is False)

        """
        if with_rel_velocity:
            nconsm = len(self.consm)

            if isinstance(m[:nconsm], list):
                m_consm = sp.Matrix(m[:nconsm])
                m_notconsm = sp.Matrix(m[nconsm:])
            else:
                m_consm = m[:nconsm]
                m_notconsm = m[nconsm:]

            return [Eq(m_consm, sp.Matrix((self.M*f)[:nconsm])),
                    *self.relative_velocity(m),
                    Eq(m_notconsm, sp.Matrix((self.Mu*f)[nconsm:]))]
        else:
            return Eq(m, self.M*f)

    def f2m(self):
        """
        Return the code expression which computes the moments from the
        distributed functions on the whole domain.
        """
        space_index = self._get_space_idx_full()
        f = self._get_indexed_on_range('f', space_index)
        m = self._get_indexed_on_range('m', space_index)
        return {'code': For(space_index, self.f2m_local(f, m))}

    def m2f_local(self, m, f, with_rel_velocity=False):
        """
        Return symbolic expression which computes the distributed functions
        from the moments.

        Parameters
        ----------

        m : SymPy Matrix
            indexed objects for the moments

        f : SymPy Matrix
            indexed objects for the distributed functions

        with_rel_velocity : boolean
            check if the scheme uses relative velocity.
            If yes, the conserved moments must be computed first.
            (default is False)

        """
        if with_rel_velocity:
            return Eq(f, self.invMu*m)
        else:
            return Eq(f, self.invM*m)

    def m2f(self):
        """
        Return the code expression which computes the distributed functions
        from the moments on the whole domain.
        """
        space_index = self._get_space_idx_full()
        f = self._get_indexed_on_range('f', space_index)
        m = self._get_indexed_on_range('m', space_index)
        return {'code': For(space_index, self.m2f_local(m, f))}

    def equilibrium_local(self, m):
        """
        Return symbolic expression which computes the equilibrium.

        Parameters
        ----------

        m : SymPy Matrix
            indexed objects for the moments

        """
        eq = self.eq.subs(list(zip(self.mv, m)))
        return Eq(m, eq)

    def equilibrium(self):
        """
        Return the code expression which computes the equilibrium
        on the whole domain.
        """
        space_index = self._get_space_idx_full()
        m = self._get_indexed_on_range('m', space_index)
        return {'code': For(space_index, self.equilibrium_local(m))}

    def relaxation_local(self, m, with_rel_velocity=False):
        """
        Return symbolic expression which computes the relaxation operator.

        Parameters
        ----------

        m : SymPy Matrix
            indexed objects for the moments

        with_rel_velocity : boolean
            check if the scheme uses relative velocity.
            (default is False)

        """
        if with_rel_velocity:
            eq = (self.Tu*self.eq).subs(list(zip(self.mv, m)))
        else:
            eq = self.eq.subs(list(zip(self.mv, m)))
        relax = (sp.ones(*self.s.shape) - self.s).multiply_elementwise(sp.Matrix(m)) + self.s.multiply_elementwise(eq)
        alltogether(relax)
        return Eq(m, relax)

    def relaxation(self):
        """
        Return the code expression which computes the relaxation
        on the whole domain.
        """
        space_index = self._get_space_idx_full()
        m = self._get_indexed_on_range('m', space_index)
        return {'code': For(space_index, self.relaxation_local(m))}

    def source_term_local(self, m):
        """
        Return symbolic expression which computes the source term
        using explicit Euler (should be more flexible in a near future).

        Parameters
        ----------

        m : SymPy Matrix
            indexed objects for the moments

        """
        rhs_eq = [mm for mm in m]
        m_local = self.settings.get('m_local', False)
        split = self.settings.get('split', False)

        indices = []
        if m_local and not split:
            local_dict = {'m': m,
                          'consm': self.consm,
                          'sorder': None,
                          'default_index': [0],
                          }
        else:
            indices_str = ['ix_', 'iy_', 'iz_']
            lbm_ind = [ix, iy, iz]
            ind_to_subs = []
            for i, sorder in enumerate(self.sorder[1:]):
                indices.append(m[0].indices[sorder])
                ind_to_subs.extend([(indices_str[i], indices[i]),
                                    (lbm_ind[i], indices[i]),
                                    ])
            local_dict = {'m': m[0].base,
                          'consm': self.consm,
                          'sorder': self.sorder,
                          'default_index': indices,
                          }
            for i, ind in enumerate(indices):
                local_dict[indices_str[i]] = ind

        for lhs, rhs in self.source_eq:
            lhs_m = lhs.subs(list(zip(self.mv, m)))
            if isinstance(rhs, (float, int)):
                rhs_m = rhs
            else:
                rhs_m = parse_expr(rhs, local_dict)
            dummy = euler(lhs_m, rhs_m)
            if indices:
                rhs_eq[lhs.i] = dummy.rhs.subs(ind_to_subs)
            else:
                rhs_eq[lhs.i] = dummy.rhs
        return [Eq(m, sp.Matrix(rhs_eq))]

    def source_term(self):
        """
        Return the code expression which computes the source terms
        on the whole inner domain.
        """
        space_index = self._get_space_idx_inner()
        m = self._get_indexed_on_range('m', space_index)
        return {'code': For(space_index, self.source_term_local(m))}

    def one_time_step_local(self, f, fnew, m):
        """
        Return symbolic expression which makes one time step of
        LBM algorithm:

            - transport
            - compute the moments from the distributed functions
            - source terms with dt/2 (with the moments)
            - relaxation (with the moments)
            - source terms with dt/2 (with the moments)
            - compute the new distributed functions from the moments

        Parameters
        ----------

        f : SymPy Matrix
            indexed objects for the old distributed functions

        fnew : SymPy Matrix
            indexed objects for the new distributed functions

        m : SymPy Matrix
            indexed objects for the moments

        """
        with_rel_velocity = True if self.rel_vel_symb else False

        code = [self.transport_local(f, fnew)]
        f2m = self.f2m_local(fnew, m, with_rel_velocity)
        if isinstance(f2m, list):
            code.extend(f2m)
        else:
            code.append(f2m)

        if self.source_eq:
            code.extend(self.source_term_local(m))

        code.append(self.relaxation_local(m, with_rel_velocity))

        if self.source_eq:
            code.extend(self.source_term_local(m))

        code.append(self.m2f_local(m, fnew, with_rel_velocity))
        return code

    def one_time_step(self):
        """
        Return the code expression which  makes one time step of
        LBM algorithm on the whole inner domain.
        """
        m_local = self.settings.get('m_local', False)
        check_isfluid = self.settings.get('check_isfluid', False)
        split = self.settings.get('split', False)

        space_index = self._get_space_idx_inner()
        if m_local:
            if split:
                m = self._get_indexed_on_range('m', space_index)
                local_vars = [m[0]]
            else:
                m = sp.MatrixSymbol('m', self.ns, 1)
                local_vars = [m]
        else:
            m = self._get_indexed_on_range('m', space_index)
            local_vars = []

        if self.rel_vel_symb:
            local_vars.extend(self.rel_vel_symb)

        f = self._get_indexed_on_velocities('f', space_index, -self.all_velocities)
        fnew = self._get_indexed_on_range('fnew', space_index)

        internal = self.one_time_step_local(f, fnew, m)

        if check_isfluid:
            valin = sp.Symbol('valin', real=True)
            in_or_out = indexed('in_or_out', [nx, ny, nz], space_index,
                                priority=self.sorder[1:])
            loop = lambda x: For(space_index, If((Eq(in_or_out, valin), x)))
        else:
            loop = lambda x: For(space_index, x)

        if split:
            # code = [loop([*self.coords(), i]) for i in internal]
            code = [loop([i]) for i in internal]
        else:
            # code = loop([*self.coords(), *internal])
            code = loop([*internal])

        return {'code': code, 'local_vars': local_vars+self.local_vars, 'settings':{"prefetch":[f[0]]}}

    @monitor
    def generate(self):
        """
        Define the routines which must be generate by code generator of SymPy
        for a given generator.
        """
        to_generate = [self.transport,
                       self.f2m,
                       self.m2f,
                       self.relaxation,
                       self.equilibrium,
                       self.one_time_step
                       ]

        if self.source_eq:
            to_generate.append(self.source_term)

        for gen in to_generate:
            name = gen.__name__
            output = gen()
            code = output['code']
            local_vars = output.get('local_vars', [])
            settings = output.get('settings', {})
            self.generator.add_routine((name, code), local_vars=local_vars, settings=settings)

    def _get_args(self, simulation, m_user=None, f_user=None, **kwargs):
        """
        Define all the arguments needed to call the generated code.
        """
        if m_user:
            mm = m_user
        else:
            mm = simulation.container.m
        m = mm.array

        dim = len(mm.nspace)
        nx = mm.nspace[0]
        X = simulation.domain.coords[0]
        if dim > 1:
            ny = mm.nspace[1]
            Y = simulation.domain.coords[1]
        if dim > 2:
            nz = mm.nspace[2]
            Z = simulation.domain.coords[1]

        if f_user:
            f = f_user.array
        else:
            f = simulation.container.F.array
        fnew = simulation.container.Fnew.array

        t = simulation.t
        dt = simulation.dt
        in_or_out = simulation.domain.in_or_out
        valin = simulation.domain.valin

        return locals()

    def call_function(self, function_name, simulation,
                      m_user=None, f_user=None, **kwargs):
        """
        Call the generated function.
        """
        from ..symbolic import call_genfunction
        func = getattr(self.generator.module, function_name)

        args = self._get_args(simulation, m_user, f_user)
        args.update(kwargs)
        call_genfunction(func, args)



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Useful functions to compute the distance
between a point and an object
"""
# pylint: disable=invalid-name

import numpy as np


def intersection_two_lines(p1, v1, p2, v2):
    """
    intersection between two lines defined by a point and a vector.
    """
    alpha = beta = None
    det = v1[1]*v2[0] - v1[0]*v2[1]
    if det != 0:
        invdet = 1./det
        c1 = p2[0] - p1[0]
        c2 = p2[1] - p1[1]
        alpha = (-v2[1]*c1 + v2[0]*c2)*invdet
        beta = (-v1[1]*c1 + v1[0]*c2)*invdet
    return alpha, beta


def distance_lines(x, y, v, p, vt, dmax, label):
    """
    return distance for several lines
    """
    v2 = np.asarray([x, y])
    alpha = 1e16*np.ones((x.size, y.size))
    border = -np.ones((x.size, y.size))
    for i, vti in enumerate(vt):
        tmp1, tmp2 = intersection_two_lines(v2, v, p[i], vti)
        if tmp1 is not None:
            if dmax is None:
                ind = np.logical_and(tmp1 > 0,
                                     np.logical_and(tmp2 >= 0, tmp2 <= 1))
            else:
                ind = np.logical_and(np.logical_and(tmp1 > 0, tmp1 <= dmax),
                                     np.logical_and(tmp2 >= 0, tmp2 <= 1))
            ind = np.where(np.logical_and(alpha > tmp1, ind))
            alpha[ind] = tmp1[ind]
            border[ind] = label[i]
    alpha[alpha == 1e16] = -1.
    return alpha, border


# pylint: disable=too-many-locals
def distance_ellipse(x, y, v, center, v1, v2, dmax, label):
    """
    return the distance according
    a line defined by a point x, y and a vector v
    to an ellipse defined by a point c and two vectors v1 and v2
    """
    # build the equation of the ellipse
    # then write the second order equation in d
    # a d**2 + b d + c = 0
    # delta = b**2-4ac
    X = x - center[0]
    Y = y - center[1]
    vx2 = v1[0]**2 + v2[0]**2
    vy2 = v1[1]**2 + v2[1]**2
    vxy = v1[0]*v1[1] + v2[0]*v2[1]
    a = v[0]**2*vy2 + v[1]**2*vx2 - 2*v[0]*v[1]*vxy
    b = 2*X*v[0]*vy2 + 2*Y*v[1]*vx2 - 2*(X*v[1]+Y*v[0])*vxy
    c = X**2*vy2 + Y**2*vx2 - 2*X*Y*vxy - (v1[0]*v2[1]-v1[1]*v2[0])**2
    delta = b**2 - 4*a*c
    ind = delta >= 0
    delta[ind] = np.sqrt(delta[ind])
    d1 = 1e16*np.ones(delta.shape)
    d2 = 1e16*np.ones(delta.shape)
    if a != 0:
        d1[ind] = (-b[ind] - delta[ind])/(2*a)
        d2[ind] = (-b[ind] + delta[ind])/(2*a)
    d1[d1 < 0] = 1e16
    d2[d2 < 0] = 1e16
    d = -np.ones(d1.shape)
    d[ind] = np.minimum(d1[ind], d2[ind])
    d[d == 1e16] = -1

    alpha = -np.ones(delta.shape)
    border = -np.ones(delta.shape)
    if dmax is None:
        ind = d > 0
    else:
        ind = np.logical_and(d > 0, d <= dmax)
    alpha[ind] = d[ind]
    border[ind] = label[0]
    return alpha, border


# pylint: disable=too-many-locals
def distance_ellipsoid(x, y, z, v, center, v1, v2, v3, dmax, label):
    """
    return the distance according
    a line defined by a point x, y, z and a vector v
    to an ellipsoid defined by a point c and three vectors v1, v2, and v3
    """
    # build the equation of the ellipsoid
    # then write the second order equation in d
    # a d**2 + b d + c = 0
    # delta = b**2-4ac
    X = x - center[0]
    Y = y - center[1]
    Z = z - center[2]
    v12 = np.cross(v1, v2)
    v23 = np.cross(v2, v3)
    v31 = np.cross(v3, v1)
    d = np.inner(v1, v23)**2
    # equation of the ellipsoid:
    # cxx XX + cyy YY + czz ZZ + cxy XY + cyz YZ + czx ZX = d
    cxx = v12[0]**2 + v23[0]**2 + v31[0]**2
    cyy = v12[1]**2 + v23[1]**2 + v31[1]**2
    czz = v12[2]**2 + v23[2]**2 + v31[2]**2
    cxy = 2 * (v12[0]*v12[1] + v23[0]*v23[1] + v31[0]*v31[1])
    cyz = 2 * (v12[1]*v12[2] + v23[1]*v23[2] + v31[1]*v31[2])
    czx = 2 * (v12[2]*v12[0] + v23[2]*v23[0] + v31[2]*v31[0])
    a = cxx*v[0]**2 + cyy*v[1]**2 + czz*v[2]**2 \
        + cxy*v[0]*v[1] + cyz*v[1]*v[2] + czx*v[2]*v[0]
    b = (2*cxx*v[0]+cxy*v[1]+czx*v[2])*X \
        + (2*cyy*v[1]+cyz*v[2]+cxy*v[0])*Y \
        + (2*czz*v[2]+czx*v[0]+cyz*v[1])*Z
    c = cxx*X**2 + cyy*Y**2 + czz*Z**2 \
        + cxy*X*Y + cyz*Y*Z + czx*Z*X - d
    delta = b**2 - 4*a*c
    ind = delta >= 0  # wird but it works
    delta[ind] = np.sqrt(delta[ind])
    d1 = 1e16*np.ones(delta.shape)
    d2 = 1e16*np.ones(delta.shape)
    d1[ind] = (-b[ind] - delta[ind])/(2*a)
    d2[ind] = (-b[ind] + delta[ind])/(2*a)
    d1[d1 < 0] = 1e16
    d2[d2 < 0] = 1e16
    d = -np.ones(d1.shape)
    d[ind] = np.minimum(d1[ind], d2[ind])
    d[d == 1e16] = -1

    alpha = -np.ones((x.size, y.size, z.size))
    border = -np.ones((x.size, y.size, z.size))
    if dmax is None:
        ind = d > 0
    else:
        ind = np.logical_and(d > 0, d <= dmax)
    alpha[ind] = d[ind]
    border[ind] = label[0]
    return alpha, border



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Cylinder element with
    - circle base
    - ellipse base
    - triangle base
    - parallelogram base
"""

# pylint: disable=invalid-name, no-member, attribute-defined-outside-init
# pylint: disable=wildcard-import, unused-wildcard-import

import logging
# from textwrap import dedent
import numpy as np

from .base import *  # pylint: disable=redefined-builtin

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Cylinder(Element):
    """
    Class Cylinder

    generic class for the cylinders
    """
    dim = 3

    def change_of_variables(self):
        """
        matrix for the change of variables
        used to write the coordinates in the basis of the cylinder
        """
        self.A = np.empty((3, 3))
        self.A[:, 0] = self.v1
        self.A[:, 1] = self.v2
        self.A[:, 2] = self.w
        self.iA = np.linalg.inv(self.A)

    def get_bounds(self):
        """
        Get the bounds of the cylinder.

        Returns
        -------

        ndarray
            minimal box where the cylinder is included

        """
        lw = [0.5*abs(self.w[k]) for k in range(len(self.w))]
        bounds_base = self.base.get_bounds()
        print(bounds_base)
        return bounds_base[0] - lw, bounds_base[1] + lw

    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the cylinder.

        Notes
        -----

        the edge of the cylinder is considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the cylinder, False otherwise)

        """
        x, y, z = grid
        xx = x - self.center[0]
        yy = y - self.center[1]
        zz = z - self.center[2]
        # the new x coordinates
        x_cyl = self.iA[0, 0]*xx + self.iA[0, 1]*yy + self.iA[0, 2]*zz
        # the new y coordinates
        y_cyl = self.iA[1, 0]*xx + self.iA[1, 1]*yy + self.iA[1, 2]*zz
        # the new z coordinates
        z_cyl = self.iA[2, 0]*xx + self.iA[2, 1]*yy + self.iA[2, 2]*zz
        return np.logical_and(
            self.base.point_inside((x_cyl, y_cyl)),
            np.abs(z_cyl) <= 1.
        )

    # pylint: disable=too-many-locals
    def distance(self, grid, v, dmax=None):
        """
        Compute the distance in the v direction between
        the cylinder and the points defined by (x, y, z).

        .. image:: ../figures/Cylinder.png
            :width: 100%

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max

        Returns
        -------

        ndarray
            array of distances

        """
        x, y, z = grid

        # rewritte the coordinates in the frame of the cylinder
        v_cyl = self.iA.dot(np.asarray(v))  # the velocity
        xx = x - self.center[0]
        yy = y - self.center[1]
        zz = z - self.center[2]
        # the new x coordinates
        x_cyl = self.iA[0, 0]*xx + self.iA[0, 1]*yy + self.iA[0, 2]*zz
        # the new y coordinates
        y_cyl = self.iA[1, 0]*xx + self.iA[1, 1]*yy + self.iA[1, 2]*zz
        # the new z coordinates
        z_cyl = self.iA[2, 0]*xx + self.iA[2, 1]*yy + self.iA[2, 2]*zz

        # considering the infinite cylinder
        alpha, border = self.base.distance(
            (x_cyl, y_cyl),
            v_cyl[:-1],
            dmax, self.label[:-2]
        )
        # indices where the intersection is too high or to low
        alpha[alpha < 0] = 1.e16
        ind = np.logical_and(alpha > 0, np.abs(z_cyl + alpha*v_cyl[2]) > 1.)
        alpha[ind] = 1.e16
        border[ind] = -1.

        # considering the two planes
        dummyf = self.base.point_inside
        if v_cyl[2] == 0:  # to avoid vertical velocities
            decal = 1.e-16
        else:
            decal = 0.
        alpha_top = (1.-z_cyl)/(v_cyl[2] + decal)
        ind = np.logical_or(
            np.logical_or(alpha_top < 0, alpha_top > dmax),
            np.logical_not(dummyf(
                (x_cyl + alpha_top*v_cyl[0], y_cyl + alpha_top*v_cyl[1])
            ))
        )
        alpha_top[ind] = 1.e16
        alpha_bot = -(1.+z_cyl)/(v_cyl[2] + decal)
        ind = np.logical_or(
            np.logical_or(alpha_bot < 0, alpha_bot > dmax),
            np.logical_not(dummyf(
                (x_cyl + alpha_bot*v_cyl[0], y_cyl + alpha_bot*v_cyl[1])
            ))
        )
        alpha_bot[ind] = 1.e16

        # considering the first intersection point
        alpha = np.amin([alpha, alpha_top, alpha_bot], axis=0)
        border[alpha == alpha_top] = self.label[-1]
        border[alpha == alpha_bot] = self.label[-2]
        alpha[alpha == 1.e16] = -1.

        return alpha, border

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('cylinder.tpl')
        elem_type = 'fluid' if self.isfluid else 'solid'
        return template.render(header=header_string(self.__class__.__name__),
                               elem=self, type=elem_type)

    # pylint: disable=too-many-locals
    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(3), alpha=1.
                  ):
        if isinstance(color, int):
            color = [color]*self.number_of_bounds
        lx_b, ly_b = self.base._visualize()  # pylint: disable=protected-access
        c = self.center
        for k in range(len(lx_b)-2):  # loop over the faces of the side
            x_b, y_b = lx_b[k], ly_b[k]
            z_b = [-1., 1.]
            X_cyl, Z_cyl = np.meshgrid(x_b, z_b)
            Y_cyl, Z_cyl = np.meshgrid(y_b, z_b)
            X = c[0] + self.A[0, 0]*X_cyl \
                + self.A[0, 1]*Y_cyl \
                + self.A[0, 2]*Z_cyl
            Y = c[1] + self.A[1, 0]*X_cyl \
                + self.A[1, 1]*Y_cyl \
                + self.A[1, 2]*Z_cyl
            Z = c[2] + self.A[2, 0]*X_cyl \
                + self.A[2, 1]*Y_cyl \
                + self.A[2, 2]*Z_cyl
            viewer.surface(X, Y, Z, color[k], alpha=alpha)
        vv = np.sin(np.linspace(0, np.pi, 10))
        Xbase = np.outer(lx_b[-2], vv)
        Ybase = np.outer(ly_b[-2], vv)
        Zbase = np.ones(Xbase.shape)
        X = c[0] + self.A[0, 0]*Xbase + self.A[0, 1]*Ybase + self.A[0, 2]*Zbase
        Y = c[1] + self.A[1, 0]*Xbase + self.A[1, 1]*Ybase + self.A[1, 2]*Zbase
        Z = c[2] + self.A[2, 0]*Xbase + self.A[2, 1]*Ybase + self.A[2, 2]*Zbase
        viewer.surface(X, Y, Z, color=color[-2], alpha=alpha)
        X = c[0] + self.A[0, 0]*Xbase + self.A[0, 1]*Ybase - self.A[0, 2]*Zbase
        Y = c[1] + self.A[1, 0]*Xbase + self.A[1, 1]*Ybase - self.A[1, 2]*Zbase
        Z = c[2] + self.A[2, 0]*Xbase + self.A[2, 1]*Ybase - self.A[2, 2]*Zbase
        viewer.surface(X, Y, Z, color=color[-2], alpha=alpha)


class CylinderCircle(Cylinder):
    """
    Class CylinderCircle

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the first vector that defines the circular section
    v2 : list
        the second vector that defines the circular section
    w : list
        the vector that defines the direction of the side
    label : list
        three integers (default [0,0,0] for the bottom, the top and the side)
    isfluid : boolean
        - True if the cylinder is added
        - False if the cylinder is deleted

    Attributes
    ----------
    number_of_bounds : int
        3
    dim: int
        3
    center : ndarray
        the coordinates of the center of the cylinder
    v1 : list
        the three coordinates of the first vector defining the base section
    v2 : list
        the three coordinates of the second vector defining the base section
    w : list
        the three coordinates of the vector defining the direction of the side
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the cylinder is added
        and False if the cylinder is deleted

    Examples
    --------

    the vertical canonical cylinder centered in (0, 0, 1/2) with radius 1

    >>> center = [0., 0., 0.5]
    >>> v1, v2 = [1., 0., 0.], [0., 1., 0.]
    >>> w = [0., 0., 1.]
    >>> CylinderCircle(center, v1, v2, w)
    +----------------+
    | CylinderCircle |
    +----------------+
        - dimension: 3
        - center: [0.  0.  0.5]
        - v1: [1. 0. 0.]
        - v2: [0. 1. 0.]
        - w: [0. 0. 1.]
        - label: [0, 0, 0]
        - type: solid

    """
    def __init__(self, center, v1, v2, w, label=0, isfluid=False):
        self.number_of_bounds = 3  # number of edges
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        self.w = np.asarray(w)
        self.change_of_variables()
        self.base = BaseCircle(self.center, self.v1, self.v2)
        Cylinder.__init__(self, label, isfluid)


class CylinderEllipse(Cylinder):
    """
    Class CylinderEllipse

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the first vector that defines the circular section
    v2 : list
        the second vector that defines the circular section
    w : list
        the vector that defines the direction of the side
    label : list
        three integers (default [0,0,0] for the bottom, the top and the side)
    isfluid : boolean
        - True if the cylinder is added
        - False if the cylinder is deleted

    Warnings
    --------

    The vectors v1 and v2 have to be orthogonal.

    Attributes
    ----------
    number_of_bounds : int
        3
    dim: int
        3
    center : ndarray
        the coordinates of the center of the cylinder
    v1 : list
        the three coordinates of the first vector defining the base section
    v2 : list
        the three coordinates of the second vector defining the base section
    w : list
        the three coordinates of the vector defining the direction of the side
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the cylinder is added
        and False if the cylinder is deleted

    Examples
    --------

    the vertical canonical cylinder centered in (0, 0, 1/2) with radius 1

    >>> center = [0., 0., 0.5]
    >>> v1, v2 = [1., 0., 0.], [0., 1., 0.]
    >>> w = [0., 0., 1.]
    >>> CylinderEllipse(center, v1, v2, w)
    +-----------------+
    | CylinderEllipse |
    +-----------------+
        - dimension: 3
        - center: [0.  0.  0.5]
        - v1: [1. 0. 0.]
        - v2: [0. 1. 0.]
        - w: [0. 0. 1.]
        - label: [0, 0, 0]
        - type: solid

    """
    def __init__(self, center, v1, v2, w, label=0, isfluid=False):
        self.number_of_bounds = 3  # number of edges
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        self.w = np.asarray(w)
        self.change_of_variables()
        self.base = BaseEllipse(self.center, self.v1, self.v2)
        Cylinder.__init__(self, label, isfluid)


class CylinderTriangle(Cylinder):
    """
    Class CylinderTriangle

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the first vector that defines the triangular section
    v2 : list
        the second vector that defines the triangular section
    w : list
        the vector that defines the direction of the side
    label : list
        three integers (default [0,0,0] for the bottom, the top and the side)
    isfluid : boolean
        - True if the cylinder is added
        - False if the cylinder is deleted

    Attributes
    ----------
    number_of_bounds : int
        5
    dim: int
        3
    center : numpy array
        the coordinates of the center of the cylinder
    v1 : list of doubles
        the three coordinates of the first vector defining the base section
    v2 : list of doubles
        the three coordinates of the second vector defining the base section
    w : list of doubles
        the three coordinates of the vector defining the direction of the side
    label : list of integers
        the list of the label of the edge
    isfluid : boolean
        True if the cylinder is added
        and False if the cylinder is deleted

    Examples
    --------

    the vertical canonical cylinder centered in (0, 0, 1/2)

    >>> center = [0., 0., 0.5]
    >>> v1, v2 = [1., 0., 0.], [0., 1., 0.]
    >>> w = [0., 0., 1.]
    >>> CylinderTriangle(center, v1, v2, w)
    +------------------+
    | CylinderTriangle |
    +------------------+
        - dimension: 3
        - center: [0.  0.  0.5]
        - v1: [1. 0. 0.]
        - v2: [0. 1. 0.]
        - w: [0. 0. 1.]
        - label: [0, 0, 0, 0, 0]
        - type: solid

    """
    def __init__(self, center, v1, v2, w, label=0, isfluid=False):
        self.number_of_bounds = 5  # number of edges
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        self.w = np.asarray(w)
        self.change_of_variables()
        self.base = BaseTriangle(self.center, self.v1, self.v2)
        Cylinder.__init__(self, label, isfluid)


class Parallelepiped(Cylinder):
    """
    Class Parallelepiped

    Parameters
    ----------
    point : list
        the three coordinates of the first point
    v0 : list
        the three coordinates of the first vector that defines the edge
    v1 : list
        the three coordinates of the second vector that defines the edge
    v2 : list
        the three coordinates of the third vector that defines the edge
    label : list
        three integers (default [0,0,0] for the bottom, the top and the side)
    isfluid : boolean
        - True if the cylinder is added
        - False if the cylinder is deleted

    Attributes
    ----------
    number_of_bounds : int
        6
    dim: int
        3
    point : ndarray
        the coordinates of the first point of the parallelepiped
    v0 : list
        the three coordinates of the first vector
    v1 : list
        the three coordinates of the second vector
    v2 : list
        the three coordinates of the third vector
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the parallelepiped is added
        and False if the parallelepiped is deleted

    Examples
    --------

    the vertical canonical cube centered in (0, 0, 0)

    >>> center = [0., 0., 0.5]
    >>> v0, v1, v2 = [1., 0., 0.], [0., 1., 0.], [0., 0., 1.]
    >>> Parallelepiped(center, v0, v1, v2)
    +----------------+
    | Parallelepiped |
    +----------------+
        - dimension: 3
        - center: [0. 0. 1.]
        - v1: [1. 0. 0.]
        - v2: [0. 1. 0.]
        - w: [0.  0.  0.5]
        - label: [0, 0, 0, 0, 0, 0]
        - type: solid

    """
    def __init__(self, point, v0, v1, v2, label=0, isfluid=False):
        self.number_of_bounds = 6  # number of edges
        self.point = np.asarray(point)
        self.v1 = np.asarray(v0)
        self.v2 = np.asarray(v1)
        self.w = .5*np.asarray(v2)
        self.center = self.point + self.w
        self.change_of_variables()
        self.base = BaseParallelogram(self.center, self.v1, self.v2)
        Cylinder.__init__(self, label, isfluid)



# FIXME: make pylint happy !
#pylint: disable=all

from sympy.core import Symbol, Expr, Tuple
from sympy.core.sympify import _sympify, sympify
from sympy.tensor import Idx, IndexedBase, Indexed
from sympy.core.basic import Basic
from sympy.core.relational import Relational
from sympy.core.compatibility import is_sequence, string_types


from sympy.core.sympify import _sympify
from sympy.functions.special.tensor_functions import KroneckerDelta
from sympy.core import Expr, Tuple, Symbol, sympify, S
from sympy.core.compatibility import is_sequence, string_types, NotIterable, range

class WithBody:
    pass
class Assignment(Relational):
    """
    Represents variable assignment for code generation.

    Parameters
    ----------
    lhs : Expr
        Sympy object representing the lhs of the expression. These should be
        singular objects, such as one would use in writing code. Notable types
        include Symbol, MatrixSymbol, MatrixElement, and Indexed. Types that
        subclass these types are also supported.

    rhs : Expr
        Sympy object representing the rhs of the expression. This can be any
        type, provided its shape corresponds to that of the lhs. For example,
        a Matrix type can be assigned to MatrixSymbol, but not to Symbol, as
        the dimensions will not align.

    Examples
    ========

    >>> from sympy import symbols, MatrixSymbol, Matrix
    >>> from sympy.printing.codeprinter import Assignment
    >>> x, y, z = symbols('x, y, z')
    >>> Assignment(x, y)
    Assignment(x, y)
    >>> Assignment(x, 0)
    Assignment(x, 0)
    >>> A = MatrixSymbol('A', 1, 3)
    >>> mat = Matrix([x, y, z]).T
    >>> Assignment(A, mat)
    Assignment(A, Matrix([[x, y, z]]))
    >>> Assignment(A[0, 1], x)
    Assignment(A[0, 1], x)
    """

    rel_op = ':='
    __slots__ = []

    def __new__(cls, lhs, rhs=0, **assumptions):
        from sympy.matrices.expressions.matexpr import (
            MatrixElement, MatrixSymbol)
        from sympy.tensor.indexed import Indexed
        lhs = _sympify(lhs)
        rhs = _sympify(rhs)
        # Tuple of things that can be on the lhs of an assignment
        assignable = (Symbol, MatrixSymbol, MatrixElement, Indexed)
        #if not isinstance(lhs, assignable):
        #    raise TypeError("Cannot assign to lhs of type %s." % type(lhs))
        # Indexed types implement shape, but don't define it until later. This
        # causes issues in assignment validation. For now, matrices are defined
        # as anything with a shape that is not an Indexed
        lhs_is_mat = hasattr(lhs, 'shape') and not isinstance(lhs, Indexed)
        rhs_is_mat = hasattr(rhs, 'shape') and not isinstance(rhs, Indexed)
        # If lhs and rhs have same structure, then this assignment is ok
        if lhs_is_mat:
            if not rhs_is_mat:
                raise ValueError("Cannot assign a scalar to a matrix.")
            elif lhs.shape != rhs.shape:
                raise ValueError("Dimensions of lhs and rhs don't align.")
        elif rhs_is_mat and not lhs_is_mat:
            raise ValueError("Cannot assign a matrix to a scalar.")
        return Relational.__new__(cls, lhs, rhs, **assumptions)

class AssignmentIf(Assignment):
    rel_op = '=='
    __slots__ = []

class For(Basic, WithBody):
    def __new__(cls, idx, expr, **kw_args):

        if isinstance(idx, Idx):
            index = Tuple(idx)
        elif is_sequence(idx):
            index = Tuple(*idx)
        else:
            raise TypeError("Loop object requires an Idx or a list of Idx.")

        if is_sequence(expr):
            expr = Tuple(*expr)
        else:
            expr = Tuple(expr)

        args = index, expr
        obj = Basic.__new__(cls, *args, **kw_args)
        return obj

    @property
    def index(self):
        return self.args[0]

    @property
    def expr(self):
        return self.args[1]

    def _sympystr(self, p):
        return p.doprint(p)

class If(Basic):
    def __new__(cls, *args, **options):
        newargs = []
        for ec in args:
            cond = ec[0]
            expr = ec[1]
            if is_sequence(expr):
                expr = Tuple(*expr)
            else:
                expr = Tuple(expr)
            newargs.append(Tuple(cond, expr))

        obj = Basic.__new__(cls, *newargs, **options)
        return obj

    @property
    def statement(self):
        return self.args

class IdxRange(Basic):
    def __new__(cls, *args):
        if len(args) > 4:
            raise ValueError("Range is defined by (top), (start, stop) or (start, stop, step)")

        # for a in args:
        #     print(type(a))
        #     if not isinstance(a, (int, Symbol, Add)):
        #         raise TypeError("args in Range must be integer or SymPy Idx")

        # expand range
        label = args[0]

        slc = slice(*args[1:])

        if slc.step == 0:
            raise ValueError("step cannot be 0")

        start, stop, step = slc.start or 0, slc.stop, slc.step or 1

        start = sympify(start)
        stop = sympify(stop)
        step = sympify(step)

        return Basic.__new__(cls, label, start, stop, step)

    @property
    def label(self):
        return self.args[0]

    @property
    def start(self):
        return self.args[1]

    @property
    def stop(self):
        return self.args[2]

    @property
    def step(self):
        return self.args[3]


class IndexedIntBase(IndexedBase):
    is_integer = True
    is_Integer = True



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Sphere element
"""

# pylint: disable=invalid-name

import logging
# from textwrap import dedent
import numpy as np

from .base import Element
from .utils import distance_ellipsoid

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Sphere(Element):
    """
    Class Sphere

    Parameters
    ----------

    center : list
        the three coordinates of the center
    radius : real
        a positive real number for the radius
    label : list
        one integer (default [0])
    isfluid : boolean
        - True if the sphere is added
        - False if the sphere is deleted

    Attributes
    ----------

    number_of_bounds : int
        1
    dim: int
        3
    center : ndarray
        the coordinates of the center of the sphere
    radius : real
        a positive real number for the radius of the sphere
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the sphere is added
        and False if the sphere is deleted

    Examples
    --------

    the sphere centered in (0, 0, 0) with radius 1

    >>> center = [0., 0., 0.]
    >>> radius = 1.
    >>> Sphere(center, radius)
    +--------+
    | Sphere |
    +--------+
        - dimension: 3
        - center: [0. 0. 0.]
        - radius: 1.0
        - label: [0]
        - type: solid

    """
    def __init__(self, center, radius, label=0, isfluid=False):
        self.number_of_bounds = 1  # number of edges
        self.dim = 3
        self.center = np.asarray(center)
        if radius >= 0:
            self.radius = radius
        else:
            log.error('The radius of the sphere should be positive')
        super(Sphere, self).__init__(label, isfluid)
        log.info(self.__str__())

    def get_bounds(self):
        """
        Get the bounds of the sphere.
        """
        return self.center - self.radius, self.center + self.radius

    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the sphere.

        Notes
        -----

        the edge of the sphere is considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the sphere, False otherwise)

        """
        x, y, z = grid
        v2 = np.asarray([
            x - self.center[0],
            y - self.center[1],
            z - self.center[2]
        ])
        return (v2[0]**2 + v2[1]**2 + v2[2]**2) <= self.radius**2

    def distance(self, grid, v, dmax=None):
        """
        Compute the distance in the v direction between
        the sphere and the points defined by (x, y, z).

        .. image:: ../figures/Sphere.png
            :width: 100%

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max

        Returns
        -------

        ndarray
            array of distances

        """
        x, y, z = grid
        v1 = self.radius*np.array([1, 0, 0])
        v2 = self.radius*np.array([0, 1, 0])
        v3 = self.radius*np.array([0, 0, 1])
        return distance_ellipsoid(
            x, y, z, v, self.center,
            v1, v2, v3, dmax, self.label
        )

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('circle.tpl')
        elem_type = 'fluid' if self.isfluid else 'solid'
        return template.render(
            header=header_string(self.__class__.__name__),
            elem=self, type=elem_type
        )

    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(3), alpha=1.
                  ):
        v1 = self.radius*np.array([1, 0, 0])*scale
        v2 = self.radius*np.array([0, 1, 0])*scale
        v3 = self.radius*np.array([0, 0, 1])*scale
        viewer.ellipse_3d(self.center*scale, v1, v2, v3, color, alpha=alpha)
        if viewlabel:
            x, y, z = self.center[0], self.center[1], self.center[2]
            viewer.text(str(self.label[0]), [x, y, z])



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause
"""
This module validates the dictionary.
"""
#pylint: disable=invalid-name
import sys
import logging
import types
from cerberus import Validator, TypeDefinition
from colorama import init, Fore, Style
import sympy

from .elements.base import Element
from .boundary import BoundaryMethod
from .algorithm import BaseAlgorithm

log = logging.getLogger(__name__) #pylint: disable=invalid-name
init(autoreset=True)

symbol_type = TypeDefinition('symbol', (sympy.Symbol, sympy.IndexedBase), ())
Validator.types_mapping['symbol'] = symbol_type

expr_type = TypeDefinition('expr', (sympy.Expr,), ())
Validator.types_mapping['expr'] = expr_type

matrix_type = TypeDefinition('matrix', (sympy.Matrix,), ())
Validator.types_mapping['matrix'] = matrix_type

element_type = TypeDefinition('element', (Element,), ())
Validator.types_mapping['element'] = element_type

function_type = TypeDefinition('function', (types.FunctionType,), ())
Validator.types_mapping['function'] = function_type

valid_prompt = lambda indent: '   | ' + ' '*indent
error_prompt = lambda indent: Fore.RED + '>>>| ' + ' '*indent + Fore.RESET
bright_error = lambda error: Style.BRIGHT +  str(error) + Style.RESET_ALL
missing_value = lambda value: '%s%s: ???%s\n'%(Fore.MAGENTA, value, Fore.RESET)

class MyValidator(Validator):
    """
    New Validator to check boundary methods.
    """
    def _validate_isboundary(self, isboundary, field, value):
        """ Test if value is a subclass of BoundaryMethod.
        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if isboundary:
            if not isinstance(value, type) or not issubclass(value, BoundaryMethod):
                self._error(field, "Must be a BoundaryMethod")

    def _validate_isalgorithm(self, isalgorithm, field, value):
        """ Test if value is a subclass of BaseAlgorithm.
        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if isalgorithm:
            if not isinstance(value, type) or not issubclass(value, BaseAlgorithm):
                self._error(field, "Must be a BaseAlgorithm")

def rec_list(mylist, errors, indent=0):
    """
    visit the list recursively and record the elements
    and the errors (if there exist) in a string.

    Parameters
    ----------

    mylist: list
        The list to print
    errors: dict
        the possible errors in this list
    indent : int
        the indentation to add for printing

    Returns
    -------

    str
        the string with the elements of the list and the errors

    """
    if isinstance(errors, str):
        return bright_error(errors) + '\n'

    s = ''
    for il, l in enumerate(mylist):
        if il not in errors.keys():
            prompt = valid_prompt
            new_errors = {}
            error_message = ''
        else:
            prompt = error_prompt
            new_errors = errors[il][0]
            error_message = bright_error(errors[il][0])

        s += prompt(indent)
        s += '%s:\n'%il
        indent += 4
        if isinstance(l, list):
            s += rec_list(l, new_errors, indent)
        elif isinstance(l, dict):
            s += rec_dict(l, new_errors, indent)
        else:
            s += prompt(indent)
            s += '%s %s\n'%(l, error_message)
        indent -= 4
    return s

def rec_dict(dico, errors, indent=0):
    """
    visit the dictionary recursively and record the elements
    and the errors (if there exist) in a string.

    Parameters
    ----------

    dico: dict
        The dictionary to print
    errors: dict
        the possible errors in this list
    indent : int
        the indentation to add for printing

    Returns
    -------

    str
        the string with the elements of the dictionary and the errors

    """
    if isinstance(errors, str):
        return bright_error(errors) + '\n'

    s = ''
    for key, value in dico.items():
        if key not in errors:
            prompt = valid_prompt
            new_errors = {}
            error_message = ''
        else:
            prompt = error_prompt
            new_errors = errors[key][0]
            error_message = bright_error(errors[key][0])

        s += prompt(indent)
        s += '%s: '%key
        indent += 4
        if isinstance(value, list) and isinstance(value[0], dict):
            s += '\n'
            s += rec_list(value, new_errors, indent)
        elif isinstance(value, dict):
            s += '\n'
            s += rec_dict(value, new_errors, indent)
        else:
            s += '%s %s\n'%(value, error_message)
        indent -= 4

    for key, value in errors.items():
        if key not in dico.keys():
            s += error_prompt(indent)
            s += missing_value(key)
    return s

def validate(dico, name):
    """
    Validation of the dictionary used in pylbm to
    instanciate Stencil, Domain, Geometry, Scheme and Simulation

    Parameters
    ----------

    dico : dict
        The dictionary to validate
    name : str
        The class name to validate ('Stencil', 'Domain', ...)

    """
    scheme = {'velocities': {'type': 'list',
                             'schema': {'type': 'integer', 'min': 0},
                             'required': name in ['Domain', 'Scheme', 'Simulation', 'Stencil']
                            },
              'M': {'type': 'matrix',
                    'required': name in ['Scheme', 'Simulation'],
                    'excludes': 'polynomials',
                    },
              'polynomials': {'type': 'list',
                              'schema': {'anyof_type': ['number', 'expr']},
                              'required': name in ['Scheme', 'Simulation'],
                              'excludes': 'M',
                             },
              'relaxation_parameters': {'type': 'list',
                                        'schema': {'anyof_type': ['number', 'expr']},
                                        'required': name in ['Scheme', 'Simulation']
                                       },
              'equilibrium': {'type': 'list',
                              'schema': {'anyof_type': ['number', 'expr']},
                              'excludes': 'feq',
                              'required': name in ['Scheme', 'Simulation']
                             },
              'feq': {'type': 'list',
                      'items': [{'type': 'function'}, {'type': 'list'}],
                      'excludes': 'equilibrium',
                      'required': name in ['Scheme', 'Simulation']
                     },
              'conserved_moments': {'anyof': [{'type': 'symbol'},
                                              {'type': 'list',
                                               'schema': {'type': 'symbol'}}
                                             ],
                                    'required': name in ['Scheme', 'Simulation']
                                   },
              'source_terms': {'type': 'dict',
                               'keyschema': {'type': 'symbol'},
                               'valueschema': {'anyof': [{'type': 'expr'},
                                                         {'type': 'number'}]},
                              },
             }

    boundary = {'method': {'type': 'dict',
                           'keyschema': {'type': 'integer'},
                           'valueschema': {'isboundary': True}
                          },
                'value': {'anyof': [{'type': 'function'},
                                    {'type': 'list',
                                     'items': [{'type': 'function'}, {'type': 'list', 'nullable': True}],
                                    }]
                         },
                'time_bc': {'type': 'boolean',
                            'default': False
                           }
               }

    simulation = {'dim': {'type': 'integer',
                          'allowed': [1, 2, 3],
                          'excludes': 'box',
                          'required': name in ['Stencil', 'Scheme']
                         },
                  'box': {'type': 'dict',
                          'schema': {'x': {'type': 'list',
                                           'items': [{'type': 'number'},
                                                     {'type': 'number'}
                                                    ]
                                          },
                                     'y': {'type': 'list',
                                           'items': [{'type': 'number'},
                                                     {'type': 'number'}
                                                    ]
                                          },
                                     'z': {'type': 'list',
                                           'items': [{'type': 'number'},
                                                     {'type': 'number'}
                                                    ]
                                          },
                                     'label': {'anyof': [{'type': 'integer'},
                                                         {'type' : 'list',
                                                          'schema': {'type': 'integer'}}
                                                        ]
                                              }
                                    },
                          'excludes': 'dim',
                          'required': name in ['Domain', 'Geometry', 'Scheme', 'Simulation', 'Stencil']
                         },
                  'elements': {'type': 'list',
                               'schema': {'type': 'element'}
                              },
                  'space_step': {'type': 'number',
                                 'min': 0,
                                 'required': name in ['Domain', 'Simulation']
                                },
                  'scheme_velocity': {'anyof_type': ['number', 'symbol'],
                                      'required': name in ['Scheme', 'Simulation']
                                     },
                  'schemes': {'type': 'list',
                              'schema': {'type': 'dict',
                                         'schema': scheme
                                        },
                              'required': name in ['Domain', 'Scheme', 'Simulation', 'Stencil']
                             },
                  'parameters': {'type': 'dict',
                                 'keyschema': {'type': 'symbol'},
                                 'valueschema': {'anyof': [{'type': 'expr'},
                                                           {'type': 'number'}]},
                                },
                  'init': {'type': 'dict',
                           'keyschema': {'type': 'symbol'},
                           'valueschema': {'anyof': [{'type': 'number'},
                                                     {'type': 'function'},
                                                     {'type': 'list',
                                                      'items': [{'type': 'function'}, {'type': 'list'}]}
                                                    ]
                                          },
                            'required': name in ['Simulation']
                            },
                  'boundary_conditions': {'type': 'dict',
                                          'keyschema': {'type': 'integer'},
                                          'valueschema': {'schema': boundary},
                                         },
                  'relative_velocity': {'type': 'list',
                                        'schema': {'anyof_type': ['number', 'expr']}
                                       },
                  'generator': {'type': 'string',
                                'allowed':['numpy', 'cython', 'loopy']
                               },
                  'codegen_dir':{'type': 'string'},
                  'lbm_algorithm': {'type': 'dict',
                                    'schema': {'name': {'isalgorithm': True},
                                               'settings': {'type': 'dict'}
                                              }
                                   },
                  'show_code': {'type': 'boolean'}
                 }

    v = MyValidator(simulation)
    is_valid = v.validate(dico)
    message = rec_dict(v.document, v.errors)

    if is_valid:
        log.info('Check the dictionary for %s class', name)
        log.info('%s', message)
    else:
        log.error('Check the dictionary for %s class', name)
        log.error(message)
        sys.exit()



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
pylbm CLI options
"""
from argparse import ArgumentParser

def options():
    """
    pylbm command line options
    """
    parser = ArgumentParser()
    logging = parser.add_argument_group('log')
    logging.add_argument("--log", dest="loglevel", default="WARNING",
                         choices=['WARNING', 'INFO', 'DEBUG', 'ERROR'],
                         help="Set the log level")
    monitoring = parser.add_argument_group('monitoring')
    monitoring.add_argument("--monitoring", action="store_true",
                            help="Set the monitoring")
    mpi = parser.add_argument_group('mpi splitting')
    mpi.add_argument("-npx", dest="npx", default=1, type=int,
                     help="Set the number of processes in x direction")
    mpi.add_argument("-npy", dest="npy", default=1, type=int,
                     help="Set the number of processes in y direction")
    mpi.add_argument("-npz", dest="npz", default=1, type=int,
                     help="Set the number of processes in z direction")
    args, _ = parser.parse_known_args()
    return args



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
stability analysis
"""

import sympy as sp
import numpy as np

from .. import viewer
from ..utils import print_progress
from ..symbolic import rel_ux, rel_uy, rel_uz, recursive_sub


class Stability:
    """
    generic class
    """
    def __init__(self, scheme):
        # pylint: disable=unsubscriptable-object
        self.nvtot = scheme.s.shape[0]
        self.consm = list(scheme.consm.keys())
        self.param = scheme.param
        self.dim = scheme.dim
        self.is_stable_l2 = True

        if scheme.rel_vel is None:
            jacobian = scheme.EQ.jacobian(self.consm)
        else:
            jacobian = (scheme.Tu * scheme.EQ).jacobian(self.consm)
        relax_mat_m = sp.eye(self.nvtot) - sp.diag(*scheme.s)
        relax_mat_m[:, :len(self.consm)] += sp.diag(*scheme.s) * jacobian

        if scheme.rel_vel is not None:
            relax_mat_m = scheme.Tmu * relax_mat_m * scheme.Tu
            relax_mat_m = relax_mat_m.subs(
                [
                    (i, j) for i, j in zip(
                        [rel_ux, rel_uy, rel_uz], scheme.rel_vel
                    )
                ]
            )
        self.relax_mat_f = scheme.invM * relax_mat_m * scheme.M
        # alltogether(self.relax_mat_f)

        velocities = sp.Matrix(scheme.stencil.get_all_velocities())
        self.velocities = np.asarray(velocities).astype('float')

        # check if we are inside a notebook
        try:
            # pylint: disable=undefined-variable
            ip = get_ipython() # noqa

            if ip.has_trait('kernel'):
                self.is_notebook = True
            else:
                self.is_notebook = False
        except:
            self.is_notebook = False

    def eigenvalues(self, consm0, n_wv):
        """
        Compute the eigenvalues of the amplification matrix
        for n_wv wave vectors
        """
        to_subs = list((i, j) for i, j in zip(self.consm, consm0))
        to_subs += list(self.param.items())
        relax_mat_f_num = recursive_sub(self.relax_mat_f, to_subs)

        if self.dim == 1:
            v_xi = np.linspace(0, 2*np.pi, n_wv, endpoint=False)
            v_xi = v_xi[np.newaxis, :]
        elif self.dim == 2:
            n_wv_0 = int(np.sqrt(n_wv))
            v_xi_0 = np.linspace(0, 2*np.pi, n_wv_0, endpoint=False)
            v_xi_x, v_xi_y = np.meshgrid(v_xi_0, v_xi_0)
            v_xi = np.array([v_xi_x.flatten(), v_xi_y.flatten()])
            n_wv = v_xi.shape[1] #pylint: disable=unsubscriptable-object
        eigs = np.empty((n_wv, self.nvtot), dtype='complex')

        if not self.is_notebook:
            print("*"*80)
            print("Compute the eigenvalues")
            print_progress(0, n_wv, barLength=50)

        relax_mat_f_num = np.asarray(relax_mat_f_num).astype('float')

        def set_matrix(wave_vector):
            return np.exp(
                self.velocities.dot(wave_vector)
            )[np.newaxis, :] * relax_mat_f_num

        for k in range(n_wv):
            data = set_matrix(1j*v_xi[:, k])
            eigs[k] = np.linalg.eig(data)[0]
            if not self.is_notebook:
                print_progress(k+1, n_wv, barLength=50)

        ind_pb, = np.where(np.max(np.abs(eigs), axis=1) > 1 + 1.e-10)
        pb_stable_l2 = v_xi[:, ind_pb]
        self.is_stable_l2 = pb_stable_l2.shape[1] == 0

        if not self.is_notebook:
            if self.is_stable_l2:
                print("*"*80)
                print("The scheme is stable")
                print("*"*80)
            else:
                print("*"*80)
                print("The scheme is not stable for these wave vectors:")
                print(pb_stable_l2.T)
                print("*"*80)

        return v_xi, eigs

    def visualize(self, dico=None, viewer_app=viewer.matplotlib_viewer):
        """
        visualize the stability
        """
        if dico is None:
            dico = {}
        consm0 = [0.] * len(self.consm)
        dicolin = dico.get('linearization', None)
        if dicolin is not None:
            for k, moment in enumerate(self.consm):
                consm0[k] = dicolin.get(moment, 0.)

        n_wv = dico.get('number_of_wave_vectors', 1024)
        v_xi, eigs = self.eigenvalues(consm0, n_wv)
        nx = v_xi.shape[1]

        fig = viewer_app.Fig(1, 2, figsize=(12.8, 6.4))  # , figsize=(12, 6))
        if self.dim == 1:
            color = 'orange'
        elif self.dim == 2:
            color = .5 + .5/np.pi*np.arctan2(v_xi[0, :], v_xi[1, :])
            color = np.repeat(
                color[np.newaxis, :], self.nvtot, axis=0
            ).flatten()

        # real and imaginary part
        view0 = fig[0]
        view0.title = "Stability: {}".format(self.is_stable_l2)
        view0.axis(-1.1, 1.1, -1.1, 1.1, aspect='equal')
        view0.grid(visible=False)
        view0.set_label('real part', 'imaginary part')
        view0.ax.set_xticks([-1, 0, 1])
        view0.ax.set_xticklabels([r"$-1$", r"$0$", r"$1$"])
        view0.ax.set_yticks([-1, 0, 1])
        view0.ax.set_yticklabels([r"$-1$", r"$0$", r"$1$"])

        theta = np.linspace(0, 2*np.pi, 1000)
        view0.plot(
            np.cos(theta), np.sin(theta),
            alpha=0.5, color='navy', width=0.5,
        )

        pos0 = np.empty((nx*self.nvtot, 2))
        for k in range(self.nvtot):
            pos0[nx*k:nx*(k+1), 0] = np.real(eigs[:, k])
            pos0[nx*k:nx*(k+1), 1] = np.imag(eigs[:, k])
        markers0 = view0.markers(pos0, 5, color=color, alpha=0.5)

        # modulus
        view1 = fig[1]
        view1.title = "Stability: {}".format(self.is_stable_l2)
        view1.axis(0, 2*np.pi, -.1, 1.1)
        view1.grid(visible=True)
        view1.set_label('wave vector modulus', 'modulus')
        view1.ax.set_xticks([k*np.pi/4 for k in range(0, 9)])
        view1.ax.set_xticklabels(
            [
                r"$0$", r"$\frac{\pi}{4}$", r"$\frac{\pi}{2}$",
                r"$\frac{3\pi}{4}$", r"$\pi$",
                r"$\frac{5\pi}{4}$", r"$\frac{3\pi}{2}$",
                r"$\frac{7\pi}{4}$", r"$2\pi$"
            ]
        )
        view1.plot(
            [0, 2*np.pi], [1., 1.],
            alpha=0.5, color='navy', width=0.5,
        )

        pos1 = np.empty((nx*self.nvtot, 2))
        for k in range(self.nvtot):
            # pos1[nx*k:nx*(k+1), 0] = np.sqrt(np.sum(v_xi**2, axis=0))
            pos1[nx*k:nx*(k+1), 0] = np.max(v_xi, axis=0)
            pos1[nx*k:nx*(k+1), 1] = np.abs(eigs[:, k])
        markers1 = view1.markers(pos1, 5, color=color, alpha=0.5)

        # create sliders to play with parameters
        dicosliders = dico.get('parameters', None)

        if self.is_notebook:
            from ipywidgets import widgets
            from IPython.display import display, clear_output

            out = widgets.Output()

            sliders = {}
            if dicosliders:
                for k, v in dicosliders.items():
                    sliders[k] = widgets.FloatSlider(value=v['init'],
                                                    min=v['range'][0],
                                                    max=v['range'][1],
                                                    step=v['step'],
                                                    continuous_update=False,
                                                    description=v.get('name', sp.pretty(k)),
                                                    layout=widgets.Layout(width='80%'))

            with out:
                fig.show()

            def update(val):  # pylint: disable=unused-argument
                for k, v in sliders.items():
                    if k in self.param.keys():
                        self.param[k] = v.value
                    for i_m, moment in enumerate(self.consm):
                        if moment == k:
                            consm0[i_m] = v.value

                v_xi, eigs = self.eigenvalues(consm0, n_wv)

                for k in range(self.nvtot):
                    pos0[nx*k:nx*(k+1), 0] = np.real(eigs[:, k])
                    pos0[nx*k:nx*(k+1), 1] = np.imag(eigs[:, k])

                markers0.set_offsets(pos0)
                view0.title = "Stability: {}".format(self.is_stable_l2)

                for k in range(self.nvtot):
                    # pos1[nx*k:nx*(k+1), 0] = np.sqrt(np.sum(v_xi**2, axis=0))
                    pos1[nx*k:nx*(k+1), 0] = np.max(v_xi, axis=0)
                    pos1[nx*k:nx*(k+1), 1] = np.abs(eigs[:, k])

                markers1.set_offsets(pos1)
                view1.title = "Stability: {}".format(self.is_stable_l2)
                fig.fig.canvas.draw_idle()
                with out:
                    clear_output(wait=True)
                    display(fig.fig)

            for k in sliders.keys():
                sliders[k].observe(update)

            display(out)
            if dicosliders:
                for k, v in dicosliders.items():
                    display(sliders[k])

        else:
            import matplotlib.pyplot as plt
            from matplotlib.widgets import Slider

            axcolor = 'lightgoldenrodyellow'

            viewer_app.Fig(figsize=(6, 2))
            sliders = {}
            item = 0
            length = 0.8/len(dicosliders)
            for k, v in dicosliders.items():
                axe = plt.axes(
                    [0.2, 0.1+item*length, 0.65, 0.8*length],
                    facecolor=axcolor,
                )
                sliders[k] = Slider(
                    axe,
                    v.get('name', sp.pretty(k)),
                    *v['range'],
                    valinit=v['init'],
                    valstep=v['step']
                )
                item += 1

            def update(val):  # pylint: disable=unused-argument
                for k, v in sliders.items():
                    if k in self.param.keys():
                        self.param[k] = v.val
                    for i_m, moment in enumerate(self.consm):
                        if moment == k:
                            consm0[i_m] = v.val

                v_xi, eigs = self.eigenvalues(consm0, n_wv)

                for k in range(self.nvtot):
                    pos0[nx*k:nx*(k+1), 0] = np.real(eigs[:, k])
                    pos0[nx*k:nx*(k+1), 1] = np.imag(eigs[:, k])

                markers0.set_offsets(pos0)
                view0.title = "Stability: {}".format(self.is_stable_l2)

                for k in range(self.nvtot):
                    # pos1[nx*k:nx*(k+1), 0] = np.sqrt(np.sum(v_xi**2, axis=0))
                    pos1[nx*k:nx*(k+1), 0] = np.max(v_xi, axis=0)
                    pos1[nx*k:nx*(k+1), 1] = np.abs(eigs[:, k])

                markers1.set_offsets(pos1)
                view1.title = "Stability: {}".format(self.is_stable_l2)
                fig.fig.canvas.draw_idle()

            for k in sliders.keys():
                sliders[k].on_changed(update)

            fig.show()



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause
# FIXME: write the documentation

# pylint: disable=missing-docstring

import logging
from six.moves import range

import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse, Polygon
from matplotlib import animation
from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=unused-import

import numpy as np

# FIXME: rewrite viewer in order to use it
#        to add multiple viewers.
# from .base import Viewer

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Fig:
    """
    Class Fig

    Parameters
    ----------

    nrows : int
        number of rows (default 1)
    ncols : int
        number of columns (default 1)
    dim : int
        dimension (default 0)
    figsize : tuple
        size of the figure (default (6, 4))

    Returns
    -------

    object :
        a figure object

    Attributes
    ----------

    fig : object
        a plot figure
    dim : int
        the dimension
    """
    def __init__(self, nrows=1, ncols=1, dim=0, figsize=(6, 4)):
        self.fig = plt.figure(figsize=figsize)
        self._grid = plt.GridSpec(nrows, ncols)
        self._plot_widgets = []
        self.dim = dim
        self.shape = (nrows, ncols)

    def fix_space(self, wspace=0.025, hspace=0.05):
        self._grid.update(wspace=wspace, hspace=hspace)

    @property
    def plot_widgets(self):
        """
        List of the associated PlotWidget instances
        """
        return tuple(self._plot_widgets)

    def __getitem__(self, idxs):
        """
        Get an axis
        """
        widget = self._grid.__getitem__(idxs)
        if self.dim < 3:
            widget = PlotWidget(
                self.fig.add_subplot(self._grid.__getitem__(idxs))
            )
        else:
            widget = PlotWidget(
                self.fig.add_subplot(
                    self._grid.__getitem__(idxs),
                    projection='3d'
                )
            )
        self._plot_widgets += [widget]
        return widget

    # pylint: disable=attribute-defined-outside-init
    def animate(self, func, interval=50):
        """
        animate the figure

        Parameters
        ----------

        func : function
            the update function
        interval : int
            the time interval between update (default 50)
        """
        self.animation = animation.FuncAnimation(
            self.fig,
            func,
            interval=interval
        )

    @staticmethod
    def show():
        plt.show()

    def close(self):
        plt.close(self.fig)


class CLine:
    """
    matplotlib object: plot
    used to allowed update function instead of set_data
    """
    def __init__(self, x, y,
                 width=2, color='black',
                 linestyle='solid', alpha=.5,
                 label=''):
        self.x = x
        self.y = y
        self.width = width
        self.color = color
        self.style = linestyle
        self.alpha = alpha
        self.label = label

    # pylint: disable=attribute-defined-outside-init
    def add(self, axe):
        self.line = axe.plot(
            self.x, self.y,
            linestyle=self.style, alpha=self.alpha,
            color=self.color, width=self.width, label=self.label
        )[0]

    def update(self, y):
        self.y = y
        self.line.set_data(self.x, self.y)


class CScatter:
    """
    matplotlib object: scatter
    used to allowed update function instead of set_offsets
    """
    def __init__(self, x, y,
                 size=5, color='black',
                 symbol='o', alpha=.5,
                 label=''):
        self.pos = np.zeros((x.size, 2))
        self.pos[:, 0] = x
        self.pos[:, 1] = y
        self.size = size
        self.color = color
        self.symbol = symbol
        self.alpha = alpha
        self.label = label

    # pylint: disable=attribute-defined-outside-init
    def add(self, axe):
        self.line = axe.markers(
            self.pos, self.size,
            symbol=self.symbol, alpha=self.alpha,
            color=self.color, label=self.label
        )

    def update(self, y):
        self.pos[:, 1] = y
        self.line.set_offsets(self.pos)


class SImage:
    """
    matplotlib object: imshow
    """
    def __init__(self, data, cmap='gist_gray', clim=(None, None), alpha=1):
        self.data = data.T
        self.cmap = cmap
        self.clim = clim
        self.alpha = alpha

    # pylint: disable=attribute-defined-outside-init
    def add(self, axe):
        self.img = axe.image(
            self.data, cmap=self.cmap, clim=self.clim,
            alpha=self.alpha,
        )

    def update(self, data):
        self.data = data.T
        self.img.set_data(self.data)


class SContour:
    """
    matplotlib object: contour
    """
    def __init__(self, data, levels=6, colors='k'):
        self.data = data.T
        self.levels = levels
        self.colors = colors

    # pylint: disable=attribute-defined-outside-init
    def add(self, axe):
        self.contour = axe.contour(
            self.data, levels=self.levels, colors=self.colors
        )
        # axe.axis_equal()
        axe.ax.clabel(self.contour, inline=1, fontsize=6)

    def update(self, data):
        # TODO
        self.data = data.T


class SScatter:
    """
    matplotlib object: 3D scatter
    """
    def __init__(self, x, y, z, size,
                 alpha=0.5, color='black', symbol='o',
                 sampling=1):
        self.size = size
        self.sampling = sampling
        self.alpha = alpha
        self.color = color
        self.symbol = symbol
        mesh_y, mesh_x = np.meshgrid(y[::self.sampling], x[::self.sampling])
        self.x = mesh_x.flatten()
        self.y = mesh_y.flatten()
        self.pos = np.zeros((self.x.size, 3))
        self.pos[:, 0] = self.x
        self.pos[:, 1] = self.y
        self.pos[:, 2] = z[::self.sampling, ::self.sampling].flatten()

    # pylint: disable=attribute-defined-outside-init
    def add(self, axe):
        self.layer = axe.markers(
            self.pos, self.size,
            alpha=self.alpha, color=self.color,
            symbol=self.symbol
        )

    def update(self, z):
        # pylint: disable=protected-access
        self.layer._offsets3d = (
            self.x, self.y,
            z[::self.sampling, ::self.sampling].flatten()
        )


# pylint: disable=too-many-public-methods
class PlotWidget:
    """
    class PlotWidget

    contains an object plotted in the figure
    """
    def __init__(self, parent):
        self.ax = parent  # pylint: disable=invalid-name

    @property
    def title(self):
        """
        get the title of the figure
        """
        return self.ax.get_title()

    @title.setter
    def title(self, text):
        """
        set the title of the figure

        Parameters
        ----------

        text : string
            the title
        """
        self.ax.set_title(text)

    def legend(self,
               loc='upper left',
               frameon=True,
               shadow=False
               ):
        """
        set the legend of the figure

        Parameters
        ----------

        loc : string
            the location (default 'upper left')
            allowed values: 'best',
            'upper left', 'upper center', 'upper right',
            'lower left', 'lower center', 'lower right'
        frameon : bool
            activate or desactivate the frame (default True)
        shadow : bool
            activate or desactivate the shadow (default False)
        """
        self.ax.legend(loc=loc, shadow=shadow, frameon=frameon)

    def text(self, text, pos,
             color='black',
             fontsize=18, fontweight='normal',
             horizontalalignment='center',
             verticalalignment='center'
             ):
        """
        add a text widget on the figure

        Parameters
        ----------

        text : string or tuple
            the text or the texts
        pos : list or tuple
            the position (2 or 3 coordinates)
        color : string or tuple
            the color of the text (default 'black')
            could be a tuple as a RGB color
        fontsize : int
            the size of the font (default 18)
        fontweight : string
            the weight of the font (default 'normal')
        horizontalalignment : string
            the horizontal alignment (default 'center')
        verticalalignment : string
            the vertical alignment (default 'center)

        Returns
        -------

        list
            the list of the text widgets
        """
        allt = []
        if isinstance(text, str):
            text = (text,)
            pos = (pos,)
        for t, p in zip(text, pos):  # pylint: disable=invalid-name
            if len(p) == 2:
                allt.append(
                    self.ax.text(
                        p[0], p[1], t,
                        color=color,
                        fontsize=fontsize,
                        fontweight=fontweight,
                        horizontalalignment=horizontalalignment,
                        verticalalignment=verticalalignment
                    )
                )
            else:
                allt.append(
                    self.ax.text(
                        p[0], p[1], p[2], t,
                        color=color,
                        fontsize=fontsize,
                        fontweight=fontweight,
                        horizontalalignment=horizontalalignment,
                        verticalalignment=verticalalignment
                    )
                )
        return allt

    def line(self, pos,
             width=2, color='black', style='solid',
             alpha=0.5, label=''
             ):
        """
        a line widget

        Parameters
        ----------

        pos : ndarray
            the coordinates of the line (2 or 3 D)

        width : int
            the width of the line (default 2)

        color : string or tuple
            the color of the line (default 'black')
            could be a tuple as a RGB color

        style : string
            the style of the line (default 'solid')

        alpha : transparency
            the transparency of the line (default 0.5)

        label : string
            the label of the line (default '')

        Returns
        -------

        object
            a graphical object

        """
        if pos.shape[1] == 2:
            return self.ax.plot(
                pos[:, 0], pos[:, 1],
                linestyle=style, alpha=alpha,
                c=color, lw=width, label=label
            )
        elif pos.shape[1] == 3:
            return self.ax.plot(
                pos[:, 0], pos[:, 1], pos[:, 2],
                linestyle=style, alpha=alpha,
                c=color, lw=width, label=label
            )
        else:
            err_msg = "Problem of dimension in viewer.line: "
            err_msg += "pos should be a ndarray of shape (N, d) "
            err_msg += "with d=2 or d=3 "
            err_msg += "(d={:d} not allowed)".format(pos.shape[1])
            log.error(err_msg)
            return

    def plot(self, x, y, z=None,
             width=2, color='black',
             marker='', linestyle='solid', alpha=1.,
             label=''
             ):
        """
        a plot widget
        """
        if z is None:
            return self.ax.plot(
                x, y, c=color,
                lw=width, marker=marker,
                label=label,
                linestyle=linestyle, alpha=alpha
            )
        else:
            return self.ax.plot(
                x, y, z, c=color,
                lw=width, marker=marker,
                label=label,
                linestyle=linestyle, alpha=alpha
            )

    # pylint: disable=unused-argument
    def segments(self, pos, width=5, color='k', alpha=1., **kwargs):
        if pos.shape[1] == 2:
            for i in range(pos.shape[0]//2):
                self.ax.plot(
                    pos[2*i:2*i+2, 0], pos[2*i:2*i+2, 1],
                    c=color, lw=width, alpha=alpha
                )
        else:
            for i in range(pos.shape[0]//2):
                self.ax.plot(
                    pos[2*i:2*i+2, 0], pos[2*i:2*i+2, 1], pos[2*i:2*i+2, 2],
                    c=color, lw=width, alpha=alpha
                )

    def clear(self):
        self.ax.clf()

    def grid(self, visible=True, which='both', alpha=1.):
        self.ax.grid(visible=visible, which=which, alpha=alpha)

    def axis(self, xmin, xmax, ymin, ymax, zmin=0, zmax=0, dim=2, aspect=None):
        if (zmin == 0 and zmax == 0) or dim <= 2:
            self.ax.set_xlim(xmin, xmax)
            self.ax.set_ylim(ymin, ymax)
            if aspect is not None:
                self.ax.set_aspect(aspect)
        else:
            self.ax.set_xlim3d(xmin, xmax)
            self.ax.set_ylim3d(ymin, ymax)
            self.ax.set_zlim3d(zmin, zmax)
        if dim == 1:
            self.ax.get_yaxis().set_visible(False)

    def axis_equal(self):
        self.ax.set_aspect('equal')

    def xaxis_set_visible(self, visible):
        self.ax.get_xaxis().set_visible(visible)

    def xaxis(self, major_ticks, minor_ticks=None):
        self.ax.set_xticks(major_ticks)
        if minor_ticks is not None:
            self.ax.set_xticks(minor_ticks, minor=True)

    def yaxis_set_visible(self, visible):
        self.ax.get_yaxis().set_visible(visible)

    def yaxis(self, major_ticks, minor_ticks=None):
        self.ax.set_yticks(major_ticks)
        if minor_ticks is not None:
            self.ax.set_yticks(minor_ticks, minor=True)

    def zaxis_set_visible(self, visible):
        self.ax.get_zaxis().set_visible(visible)

    def zaxis(self, major_ticks, minor_ticks=None):
        self.ax.set_zticks(major_ticks)
        if minor_ticks is not None:
            self.ax.set_zticks(minor_ticks, minor=True)

    def set_label(self, xlab, ylab, zlab=None):
        if xlab is not None:
            self.ax.set_xlabel(xlab)
        if ylab is not None:
            self.ax.set_ylabel(ylab)
        if zlab is not None:
            self.ax.set_zlabel(zlab)

    def ellipse(self, pos, radius, color, angle=0., alpha=1):
        return self.ax.add_patch(
            Ellipse(
                xy=pos,
                width=2*radius[0], height=2*radius[1],
                angle=angle*180/np.pi,
                fill=True, color=color, alpha=alpha
            )
        )

    def polygon(self, pos, color, alpha=1.):
        return self.ax.add_patch(
            Polygon(
                pos, closed=True, fill=True, color=color, alpha=alpha
            )
        )

    def surface(self, x, y, z, color, alpha=0.5):
        return self.ax.plot_surface(
            x, y, z,
            rstride=1, cstride=1, color=color,
            shade=False, alpha=alpha,
            antialiased=False, linewidth=.5
        )

    # pylint: disable=invalid-name, too-many-locals
    def ellipse_3d(self, pos, a, b, c, color, alpha=1.):
        u = np.linspace(0, 2.*np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        CS = np.outer(np.cos(u), np.sin(v))
        SS = np.outer(np.sin(u), np.sin(v))
        C = np.outer(np.ones(np.size(u)), np.cos(v))
        x = pos[0] + a[0]*CS + b[0]*SS + c[0]*C
        y = pos[1] + a[1]*CS + b[1]*SS + c[1]*C
        z = pos[2] + a[2]*CS + b[2]*SS + c[2]*C
        return self.ax.plot_surface(
            x, y, z,
            rstride=4, cstride=4,
            color=color[0], alpha=alpha
        )

    def markers(self, pos, size, color='k',
                symbol='o', alpha=1., dim=None, label=''
                ):
        if dim is None:
            dim = pos.shape[1]
        if pos.shape[1] == 2 or dim <= 2:
            return self.ax.scatter(
                pos[:, 0], pos[:, 1],
                s=size, c=color, marker=symbol,
                alpha=alpha, label=label
            )
        else:
            posx, posy, posz = pos[:, 0], pos[:, 1], pos[:, 2]
            return self.ax.scatter(
                posx, posy, posz,
                c=color, marker=symbol, s=size,
                alpha=alpha, label=label
            )

    def image(self, f, fargs=(), cmap='gist_gray', clim=(None, None), alpha=1):
        if isinstance(f, np.ndarray):
            data = f
        else:
            data = f(*fargs)
        image = self.ax.imshow(
            data, origin='lower',
            vmin=clim[0], vmax=clim[1],
            cmap=cmap, interpolation='nearest',
            alpha=alpha,
        )
        return image

    def contour(self, Z, levels=10, colors='k'):
        contour = self.ax.contour(Z, levels=levels, colors=colors)
        return contour


    @staticmethod
    def draw():
        plt.show()

    @property
    def is3d(self):
        return False

    def CurveLine(self, x, y,
                  width=2, color='black',
                  linestyle='solid', alpha=.5,
                  label=''):
        line = CLine(
            x, y,
            width=width, color=color,
            linestyle=linestyle, alpha=alpha,
            label=label
        )
        line.add(self)
        return line

    def CurveScatter(self, x, y, size=5, color='black',
                     symbol='o', alpha=.5,
                     dim=None, label=''
                     ):
        line = CScatter(
            x, y,
            size=size, color=color,
            symbol=symbol, alpha=alpha,
            label=label
        )
        line.add(self)
        return line

    def SurfaceImage(self, data, cmap='gist_gray', clim=(None, None), alpha=1):
        layer = SImage(data, cmap=cmap, clim=clim, alpha=alpha)
        layer.add(self)
        return layer

    def SurfaceContour(self, data, levels=6, colors='k'):
        layer = SContour(data, levels=levels, colors=colors)
        layer.add(self)
        return layer

    def SurfaceScatter(self, x, y, z, size, sampling=1,
                       alpha=0.5, color='black', symbol='o'
                       ):
        layer = SScatter(
            x, y, z, size, sampling=sampling,
            alpha=alpha, color=color, symbol=symbol
        )
        layer.add(self)
        return layer



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
pylbm simulation
"""

import os
import sys
import logging
import types
import numpy as np
from sympy.parsing.sympy_parser import parse_expr
import mpi4py.MPI as mpi

from .domain import Domain
from .scheme import Scheme
from .boundary import Boundary
from . import utils
from .validator import validate
from .context import set_queue
from .generator import Generator
from .container import NumpyContainer, CythonContainer, LoopyContainer
from .algorithm import PullAlgorithm
from .monitoring import Monitor, monitor

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Simulation:
    """
    create a class simulation

    Parameters
    ----------

    dico : dictionary
    domain : object of class :py:class:`Domain<pylbm.domain.Domain>`, optional
    scheme : object of class :py:class:`Scheme<pylbm.scheme.Scheme>`, optional
    type :   optional argument (default value is 'float64')

    Attributes
    ----------

    dim : int
      spatial dimension
    type : float64
      the type of the values
    domain : :py:class:`Domain<pylbm.domain.Domain>`
      the domain given in argument
    scheme : :py:class:`Scheme<pylbm.scheme.Scheme>`
      the scheme given in argument
    m : numpy array
      a numpy array that contains the values of the moments in each point
    F : numpy array
      a numpy array that contains the values of the distribution functions
      in each point
    m_halo : numpy array
      a numpy array that contains the values of the moments in each point
    F_halo : numpy array
      a numpy array that contains the values of the distribution functions
      in each point

    Examples
    --------

    see demo/examples/

    Notes
    -----

    The methods
    :py:meth:`transport<pylbm.simulation.Simulation.transport>`,
    :py:meth:`relaxation<pylbm.simulation.Simulation.relaxation>`,
    :py:meth:`equilibrium<pylbm.simulation.Simulation.equilibrium>`,
    :py:meth:`f2m<pylbm.simulation.Simulation.f2m>`,
    :py:meth:`m2f<pylbm.simulation.Simulation.m2f>`,
    :py:meth:`boundary_condition<pylbm.simulation.Simulation.boundary_condition>`,
    and
    :py:meth:`one_time_step<pylbm.simulation.Simulation.one_time_step>`
    are just call of the methods of the class
    :py:class:`Scheme<pylbm.scheme.Scheme>`.
    """
    # pylint: disable=too-many-branches, too-many-statements, too-many-locals
    def __init__(self, dico,
                 sorder=None, dtype='float64',
                 check_inverse=False
                 ):
        validate(dico, __class__.__name__) #pylint: disable=undefined-variable

        self.domain = Domain(dico, need_validation=False)
        domain_size = mpi.COMM_WORLD.allreduce(sendobj=np.prod(self.domain.shape_in))
        Monitor.set_size(domain_size)

        self.scheme = Scheme(dico, check_inverse=check_inverse, need_validation=False)
        if self.domain.dim != self.scheme.dim:
            log.error('Solution: the dimension of the domain and of the scheme are not the same\n')
            sys.exit()

        self._update_m = True
        self.t = 0.
        self.nt = 0
        self.dt = self.domain.dx/self.scheme.la
        self.dim = self.domain.dim

        codegen_dir = dico.get('codegen_dir', None)
        if codegen_dir:
            codegen_dir = os.path.realpath(codegen_dir)

        self.generator = Generator(dico.get('generator', "CYTHON").upper(),
                                   codegen_dir,
                                   dico.get('show_code', False))

        # FIXME remove that !!
        set_queue(self.generator.backend)

        self.container = self._get_container(sorder)
        if self.container.gpu_support:
            self.domain.in_or_out = self.container.move2gpu(self.domain.in_or_out)
            self.container.F.generate(self.generator)
            self.container.Fnew.generate(self.generator)
        sorder = self.container.sorder

        # Generate the numerical code for the LBM and for the boundary conditions
        self.algo = self._get_algorithm(dico, sorder)
        self.algo.generate()

        self.bc = Boundary(self.domain, self.generator, dico)
        for method in self.bc.methods:
            method.set_iload()
            method.generate(self.container.sorder)

        self.generator.compile()

        # Initialize the solution and the rhs of boundary conditions
        self.initialization(dico)
        for method in self.bc.methods:
            method.prepare_rhs(self)
            method.fix_iload()
            method.set_rhs()
            method.move2gpu()

        log.info(self.__str__())

    def _get_container(self, sorder):
        container_type = {'NUMPY': NumpyContainer,
                          'CYTHON': CythonContainer,
                          'LOOPY': LoopyContainer
        }
        return container_type[self.generator.backend](self.domain, self.scheme, sorder)

    def _get_default_algo_settings(self):
        if self.generator.backend == 'NUMPY':
            return {'m_local': False, 'split': False, 'check_isfluid': False}
        else:
            return {'m_local': True, 'split': False, 'check_isfluid': False}

    def _get_algorithm(self, dico, sorder):
        algo_method = PullAlgorithm
        user_settings = {}
        dummy = dico.get('lbm_algorithm', None)
        if dummy:
            algo_method = dummy.get('name', PullAlgorithm)
            user_settings = dummy.get('settings', {})
        algo_settings = self._get_default_algo_settings()
        algo_settings.update(user_settings)

        return algo_method(self.scheme, sorder, self.generator, algo_settings)

    @utils.itemproperty
    def m_halo(self, i):
        """
        get the moment i on the whole domain with halo points.
        """
        if self._update_m:
            self._update_m = False
            self.f2m()
        return self.container.m[i]

    @m_halo.setter
    def m_halo(self, i, value):
        self._update_m = False
        self.container.m[i] = value

    @utils.itemproperty
    def m(self, i):
        """
        get the moment i in the interior domain.
        """
        if self._update_m:
            self._update_m = False
            self.f2m()

        return self.container.m._in(i) #pylint: disable=protected-access

    @utils.itemproperty
    def F_halo(self, i):
        """
        get the distribution function i on the whole domain with halo points.
        """
        return self.container.F[i]

    @F_halo.setter
    def F_halo(self, i, value):
        self._update_m = True
        self.container.F[i] = value

    @utils.itemproperty
    def F(self, i):
        """
        get the distribution function i in the interior domain.
        """
        return self.container.F._in(i) #pylint: disable=protected-access

    def __str__(self):
        from .utils import header_string
        from .jinja_env import env
        template = env.get_template('simulation.tpl')
        return template.render(header=header_string("Simulation information"),
                               simu=self)

    def __repr__(self):
        return self.__str__()

    @monitor
    def initialization(self, dico):
        """
        initialize all the numy array with the initial conditions
        set the initial values to the numpy arrays _F and _m

        Parameters
        ----------

        dico : the dictionary with the `key:value` 'init'


        Notes
        -----

        The initial values are set to _m, the array _F is then initialized
        with the equilibrium values.
        If the initial values have to be set to _F, use the optional
        `key:value` 'inittype' with the value 'distributions'
        (default value is set to 'moments').

        """
        # type of initialization
        # by default, the initialization is on the moments
        # else, it could be distributions
        inittype = dico.get('inittype', 'moments')
        coords = np.meshgrid(*(c for c in self.domain.coords_halo), sparse=True, indexing='ij')

        if inittype == 'moments':
            array_to_init = self.container.m
        elif inittype == 'distributions':
            array_to_init = self.container.F
        else:
            sss = 'Error in the creation of the scheme: wrong dictionnary\n'
            sss += 'the key `inittype` should be moments or distributions'
            log.error(sss)
            sys.exit()

        init_data = dico.get('init', None)
        if init_data is None:
            log.warning("You don't define initialization step for your conserved moments")
            return

        for k, v in init_data.items():
            if isinstance(v, tuple):
                f = v[0]
                extraargs = v[1] if len(v) == 2 else ()
                fargs = tuple(coords) + extraargs
                array_to_init[k] = f(*fargs)
            elif isinstance(v, types.FunctionType):
                fargs = tuple(coords)
                array_to_init[k] = v(*fargs)
            else:
                array_to_init[k] = v

        if inittype == 'moments':
            self.equilibrium()
            self.m2f()
        elif inittype == 'distributions':
            self.f2m()

        self.container.Fnew.array[:] = self.container.F.array[:]

    def transport(self, **kwargs):
        """
        compute the transport phase on distribution functions
        (the array _F is modified)
        """
        self.algo.call_function('transport', self, **kwargs)

    def relaxation(self, **kwargs):
        """
        compute the relaxation phase on moments
        (the array _m is modified)
        """
        self.algo.call_function('relaxation', self, **kwargs)

    def source_term(self, fraction_of_time_step=1., **kwargs):
        """
        compute the source term phase on moments
        (the array _m is modified)
        """
        self.algo.call_function('source_term', self, **kwargs)

    @monitor
    def f2m(self, **kwargs):
        """
        compute the moments from the distribution functions
        (the array _m is modified)
        """
        self.algo.call_function('f2m', self, **kwargs)

    @monitor
    def m2f(self, m_user=None, f_user=None, **kwargs):
        """
        compute the distribution functions from the moments
        (the array _F is modified)
        """
        self.algo.call_function('m2f', self, m_user, f_user, **kwargs)

    @monitor
    def equilibrium(self, m_user=None, **kwargs):
        """
        set the moments to the equilibrium values
        (the array _m is modified)

        Notes
        -----

        Another moments vector can be set to equilibrium values:
        use directly the method of the class Scheme
        """
        self.algo.call_function('equilibrium', self, m_user, **kwargs)

    @monitor
    def boundary_condition(self, **kwargs):
        """
        perform the boundary conditions

        Notes
        -----

        The array _F is modified in the phantom array (outer points)
        according to the specified boundary conditions.
        """
        f = self.container.F
        f.update()

        for method in self.bc.methods:
            method.update_feq(self)
            method.set_rhs()
            method.update(f, **kwargs)

    @monitor
    def one_time_step(self, **kwargs):
        """
        compute one time step

        Notes
        -----

        Modify the arrays _F and _m in order to go further of dt.
        This function is equivalent to successively use

        - boundary_condition
        - transport
        - f2m
        - relaxation
        - m2f
        """
        self._update_m = True # we recompute f so m will be not correct

        self.boundary_condition(**kwargs)

        self.algo.call_function('one_time_step', self, **kwargs)
        self.container.F, self.container.Fnew = self.container.Fnew, self.container.F

        self.t += self.dt
        self.nt += 1



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause
"""
This module provides the classes to manage the velocities of
lattice Boltzmann schemes.
"""

from math import sqrt
import logging
import numpy as np

from .utils import itemproperty
from .geometry import get_box
from . import viewer
from .validator import validate

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def cmp(a, b):
    """
    cmp function like in python 2.
    """
    return (a > b) - (a < b)


def permute_in_place(iterable):
    """
    Function that returns an iterator of all the permutations of a list

    Parameters
    ----------

    iterable: list

    Yield
    -----

    list
        successive permutations of elements in the list iterable

    Note
    ----

    The set of all the permutations is not created in the memory,
    so it can just be used in a loop.

    It can be used as permutations() of the itertools package but avoids the
    multiple occurrences of a same output list.

    Examples
    --------

    .. code::python
        >>> import itertools
        >>> for k in itertools.permutations([0, 0, 1]):
        ...     print(k)
        ...
        (0, 0, 1)
        (0, 1, 0)
        (0, 0, 1)
        (0, 1, 0)
        (1, 0, 0)
        (1, 0, 0)

        >>> for k in permute_in_place([0, 0, 1]):
        ...     print(k)
        ...
        [0, 0, 1]
        [0, 1, 0]
        [1, 0, 0]
    """
    iterable.sort()
    yield list(iterable)

    if len(iterable) <= 1:
        return

    first = 0
    last = len(iterable)
    while 1:
        i = last - 1

        while 1:
            i = i - 1
            if iterable[i] < iterable[i + 1]:
                j = last - 1
                while not iterable[i] < iterable[j]:
                    j = j - 1
                # swap the values
                iterable[i], iterable[j] = iterable[j], iterable[i]
                part = iterable[i + 1:last]
                part.reverse()
                iterable[i + 1:last] = part
                yield list(iterable)
                break
            if i == first:
                iterable.reverse()
                return


class Singleton(type):
    """
    Singleton metaclasss
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        key = (cls, args, str(kwargs))
        if key not in cls._instances:
            cls._instances[key] = super(
                Singleton, cls
            ).__call__(*args, **kwargs)
        return cls._instances[key]


class Velocity:
    """
    Create a velocity.

    Parameters
    ----------

    dim : int, optional
         The dimension of the velocity.
    num : int, optional
         The number of the velocity in the numbering convention
         of Lattice-Boltzmann scheme.
    vx : int, optional
         The x component of the velocity vector.
    vy : int, optional
         The y component of the velocity vector.
    vz : int, optional
         The z component of the velocity vector.

    Attributes
    ----------

    dim : int
        The dimension of the velocity.
    num
        The number of the velocity in the numbering convention
        of Lattice-Boltzmann scheme.
    vx : int
        The x component of the velocity vector.
    vy : int
        The y component of the velocity vector.
    vz : int
        The z component of the velocity vector.
    v : list

    Examples
    --------

    Create a velocity with the dimension and the number

    >>> v = Velocity(dim=1, num=2)
    >>> v
    (2: -1)

    Create a velocity with a direction

    >>> v = Velocity(vx=1, vy=1)
    >>> v
    (5: 1, 1)

    Notes
    ------

    .. plot:: codes/Velocities.py

    """
    __metaclass__ = Singleton
    _d = 1e3
    _R2 = np.array([[[5, 6, 4], [_d, _d, 2], [2, 5, 3]],
                    [[3, _d, _d], [_d, -1, _d], [1, _d, _d]],
                    [[6, 7, 7], [_d, _d, 0], [1, 4, 0]]], dtype=np.int)

    def __init__(self, dim=None, num=None, vx=None, vy=None, vz=None):
        self.dim = dim
        self.num = num

        self.vx = vx
        self.vy = vy
        self.vz = vz

        if dim is None:
            if vz is not None:
                self.dim = 3
            elif vy is not None:
                self.dim = 2
            elif vx is not None:
                self.dim = 1
            else:
                err_msg = "The parameters could not be all None "
                err_msg += "when creating a velocity"
                log.error(err_msg)

        if num is None:
            self._set_num()
        if vx is None:
            self._set_coord()

    @property
    def v(self):
        """
        velocity
        """
        return [self.vx, self.vy, self.vz][:self.dim]

    @property
    def v_full(self):
        """
        velocity filled with 0 in 1d and 2d
        """
        v_filled = [0]*3
        v_filled[:self.dim] = self.v
        return v_filled

    def __str__(self):
        output = '({:d}: {:d}'.format(self.num, self.vx)
        if self.vy is not None:
            output += ', {:d}'.format(self.vy)
        if self.vz is not None:
            output += ', {:d}'.format(self.vz)
        output += ')'
        return output

    def __repr__(self):
        return self.__str__()

    # pylint: disable=invalid-unary-operand-type
    def get_symmetric(self, axis=None):
        """
        return the symmetric velocity.

        Parameters
        ----------

        axis : the axis of the symmetry, optional
               (None involves the symmetric with the origin,
               0 with the x axis, 1 with the y axis, and 2 with the z axis)

        Returns
        -------

        Velocity
            The symmetric of the velocity

        Raises
        ------

        ValueError
            if axis is not None and axis < 0 or axis >= dim

        """
        if axis is not None and (axis >= self.dim or axis < 0):
            err_msg = "axis must be less than the dimension of the "
            err_msg += "velocity (axis={:d}, dim={:d})".format(axis, self.dim)
            log.error(err_msg)
            raise ValueError

        svx = -self.vx
        svy = None if self.vy is None else -self.vy
        svz = None if self.vz is None else -self.vz

        if axis is None:
            return Velocity(vx=svx, vy=svy, vz=svz)
        if axis == 0:
            return Velocity(vx=self.vx, vy=svy, vz=svz)
        if axis == 1:
            return Velocity(vx=svx, vy=self.vy, vz=svz)
        return Velocity(vx=svx, vy=svy, vz=self.vz)

    # pylint: disable=invalid-name, too-many-locals
    # pylint: disable=too-many-nested-blocks, too-complex
    def _set_num(self):
        # computes the number of the velocity
        # the coordinates being given
        if self.dim == 1:
            avx = abs(self.vx)
            self.num = (2*avx) - (1 if self.vx > 0 else 0)
            return
        elif self.dim == 2:
            avx = abs(self.vx)
            avy = abs(self.vy)
            T1 = cmp(self.vx, 0)
            T2 = cmp(self.vy, 0)
            T3 = cmp(avx, avy)
            p = (2*max(avx, avy) - 1)
            p *= p
            q = 8*min(avx, avy)*abs(T3)
            r = self._R2[T1 + 1, T2 + 1, T3 + 1]
            self.num = int(p + q + r)
            return
        elif self.dim == 3:
            count = 0
            sign = [1, -1]
            for k in range(100):
                for i in range(k + 1):
                    for j in range(i + 1):
                        for (kk, ii, jj) in permute_in_place([k, i, j]):
                            # loop over + and - if kk > 0
                            for pmk in sign[0: kk + 1]:
                                # loop over + and - if ii > 0
                                for pmi in sign[0:ii + 1]:
                                    # loop over + and - if jj > 0
                                    for pmj in sign[0:jj + 1]:
                                        if self.vx == pmk*kk and \
                                           self.vy == pmi*ii and \
                                           self.vz == pmj*jj:
                                            self.num = count
                                            return
                                        else:
                                            count += 1
        log.error("The number of the velocity %s is not found", self.__str__())

    # pylint: disable=too-many-locals, too-many-branches, too-complex
    def _set_coord(self):
        # computes the coordinates of the velocity
        # the number being given
        if self.dim == 1:
            n = self.num + 1
            self.vx = int((1 - 2*(n % 2))*(n/2))
            return
        elif self.dim == 2:
            n = (int)((sqrt(self.num)+1)/2)
            p = self.num - (2*n-1)**2
            if p < 4:
                Lx, Ly = [n, 0, -n, 0], [0, n, 0, -n]
                vx, vy = Lx[p], Ly[p]
            elif p < 8:
                Lx, Ly = [n, -n, -n, n], [n, n, -n, -n]
                vx, vy = Lx[p-4], Ly[p-4]
            else:
                k, l = n, p/8
                Lx = [k, l, -l, -k, -k, -l, l, k]
                Ly = [l, k, k, l, -l, -k, -k, -l]
                vx, vy = Lx[p % 8], Ly[p % 8]
            self.vx = int(vx)
            self.vy = int(vy)
            return
        elif self.dim == 3:
            count = 0
            sign = [1, -1]
            for k in range(100):
                for i in range(k + 1):
                    for j in range(i + 1):
                        for (kk, ii, jj) in permute_in_place([k, i, j]):
                            # loop over + and - if kk > 0
                            for pmk in sign[0:kk + 1]:
                                # loop over + and - if ii > 0
                                for pmi in sign[0:ii + 1]:
                                    # loop over + and - if jj > 0
                                    for pmj in sign[0:jj + 1]:
                                        if self.num == count:
                                            self.vx = int(pmk*kk)
                                            self.vy = int(pmi*ii)
                                            self.vz = int(pmj*jj)
                                            return
                                        else:
                                            count += 1
        log.error("The velocity number %d cannot be computed", self.num)


class OneStencil:
    """
    Create a stencil of a LBM scheme.

    Parameters
    ----------

    v : list
      the list of the velocities of that stencil
    nv : int
      size of the list
    num2index : list of integers
      link between the velocity number and its position in the unique
      velocities array

    Attributes
    ----------

    v : list
      the list of the velocities of that stencil
    nv : int
      size of the list v
    num
      the numbering of the velocities
    vx
      the x component of the velocities
    vy
      the y component of the velocities
    vz
    """

    def __init__(self, v, nv):
        self.v = v
        self.nv = nv

    @property
    def num(self):
        """
        the numbering of the velocities.
        """
        vectorize = np.vectorize(lambda obj: obj.num)
        return vectorize(self.v)

    @property
    def vx(self):
        """
        the x component of the velocities.
        """
        vectorize = np.vectorize(lambda obj: obj.vx)
        return vectorize(self.v)

    @property
    def vy(self):
        """
        the y component of the velocities.
        """
        vectorize = np.vectorize(lambda obj: obj.vy)
        return vectorize(self.v)

    @property
    def vz(self):
        """
        the z component of the velocities.
        """
        vectorize = np.vectorize(lambda obj: obj.vz)
        return vectorize(self.v)


class Stencil(list):
    """
    Create the stencil of velocities used by the scheme(s).

    The numbering of the velocities follows the convention for 1D and 2D.

    Parameters
    ----------

    dico : a dictionary that contains the following `key:value`
        - dim : the value of the spatial dimension (1, 2 or 3)
        - schemes : a list of dictionaries that contain
                    the key:value velocities

        .. code-block:: python

            [
                {
                    'velocities': [...]
                },
                {
                    'velocities': [...]
                },
                {
                    'velocities': [...]
                },
                ...
            ]

    Attributes
    ----------

    dim : int
        the spatial dimension (1, 2 or 3).
    unique_velocities : NumPy array
        array of all velocities involved in the stencils.
        Each unique velocity appeared only once.
    uvx : NumPy array
        the x component of the unique velocities.
    uvy : NumPy array
        the y component of the unique velocities.
    uvz : NumPy array
        the z component of the unique velocities.
    unum : NumPy array
        the numbering of the unique velocities.
    vmax : int
        the maximal velocity in norm for each spatial direction.
    vmin : int
        the minimal velocity in norm for each spatial direction.
    vmax_full : int
        the maximal velocity in norm for the 3 spatial direction
        even in dim 1 or 2.
    nstencils : int
        the number of elementary stencils.
    nv : list of integers
        the number of velocities for each elementary stencil.
    v : list of velocities
        list of all the velocities for each elementary stencil.
    vx : NumPy array
        the x component of the velocities for the stencil k.
    vy : NumPy array
        the y component of the velocities for the stencil k.
    vz : NumPy array
        the z component of the velocities for the stencil k.
    num : NumPy array
        the numbering of the velocities for the stencil k.
    nv_ptr : list of integers
        used to obtain the list of the velocities involved in a stencil.
        For instance, the list for the kth stencil is
        v[nv_ptr[k]:nv_ptr[k+1]]
    unvtot : int
        the number of unique velocities involved in the stencils.

    Notes
    -----

    The velocities for each schemes are defined as a Python list.

    Examples
    --------

    >>> s = Stencil({'dim': 1,
    ...              'schemes':[{'velocities': list(range(9))},],
    ...             })
    >>> s
    +---------------------+
    | Stencil information |
    +---------------------+
        - spatial dimension: 1
        - minimal velocity in each direction: [-4]
        - maximal velocity in each direction: [4]
        - information for each elementary stencil:
            stencil 0
                - number of velocities: 9
                - velocities
                    (0: 0)
                    (1: 1)
                    (2: -1)
                    (3: 2)
                    (4: -2)
                    (5: 3)
                    (6: -3)
                    (7: 4)
                    (8: -4)
    >>> s = Stencil({'dim': 2,
    ...              'schemes':[
                                   {'velocities': list(range(9))},
    ...                            {'velocities': list(range(49))},
    ...                        ],
    ...            })
    >>> s
    +---------------------+
    | Stencil information |
    +---------------------+
        - spatial dimension: 2
        - minimal velocity in each direction: [-3 -3]
        - maximal velocity in each direction: [3 3]
        - information for each elementary stencil:
            stencil 0
                - number of velocities: 9
                - velocities
                    (0: 0, 0)
                    (1: 1, 0)
                    (2: 0, 1)
                    (3: -1, 0)
                    (4: 0, -1)
                    (5: 1, 1)
                    (6: -1, 1)
                    (7: -1, -1)
                    (8: 1, -1)
            stencil 1
                - number of velocities: 49
                - velocities
                    (0: 0, 0)
                    (1: 1, 0)
                    (2: 0, 1)
                    (3: -1, 0)
                    (4: 0, -1)
                    (5: 1, 1)
                    (6: -1, 1)
                    (7: -1, -1)
                    (8: 1, -1)
                    (9: 2, 0)
                    (10: 0, 2)
                    (11: -2, 0)
                    (12: 0, -2)
                    (13: 2, 2)
                    (14: -2, 2)
                    (15: -2, -2)
                    (16: 2, -2)
                    (17: 2, 1)
                    (18: 1, 2)
                    (19: -1, 2)
                    (20: -2, 1)
                    (21: -2, -1)
                    (22: -1, -2)
                    (23: 1, -2)
                    (24: 2, -1)
                    (25: 3, 0)
                    (26: 0, 3)
                    (27: -3, 0)
                    (28: 0, -3)
                    (29: 3, 3)
                    (30: -3, 3)
                    (31: -3, -3)
                    (32: 3, -3)
                    (33: 3, 1)
                    (34: 1, 3)
                    (35: -1, 3)
                    (36: -3, 1)
                    (37: -3, -1)
                    (38: -1, -3)
                    (39: 1, -3)
                    (40: 3, -1)
                    (41: 3, 2)
                    (42: 2, 3)
                    (43: -2, 3)
                    (44: -3, 2)
                    (45: -3, -2)
                    (46: -2, -3)
                    (47: 2, -3)
                    (48: 3, -2)

    get the x component of the unique velocities

    >>> s.uvx
    array([ 0,  1,  0, -1,  0,  1, -1, -1,  1,  2,  0, -2,  0,  2, -2, -2,  2,
            2,  1, -1, -2, -2, -1,  1,  2,  3,  0, -3,  0,  3, -3, -3,  3,  3,
            1, -1, -3, -3, -1,  1,  3,  3,  2, -2, -3, -3, -2,  2,  3])

    get the y component of the velocity for the second stencil

    >>> s.vy[1]
    array([ 0,  0,  1,  0, -1,  1,  1, -1, -1,  0,  2,  0, -2,  2,  2, -2, -2,
            1,  2,  2,  1, -1, -2, -2, -1,  0,  3,  0, -3,  3,  3, -3, -3,  1,
            3,  3,  1, -1, -3, -3, -1,  2,  3,  3,  2, -2, -3, -3, -2])

    """
    def __init__(self, dico, need_validation=True):
        super(Stencil, self).__init__()

        if need_validation:
            validate(dico, __class__.__name__) #pylint: disable=undefined-variable

        # get the dimension of the stencil
        # (given in the dictionnary or computed
        #  through the geometrical box)
        self.dim = self.extract_dim(dico)

        # get the schemes
        schemes_velocities = []
        schemes = dico['schemes']
        for scheme in schemes:
            schemes_velocities.append(np.asarray(scheme['velocities']))
        self.nstencils = len(schemes_velocities)

        # get the unique velocities involved in the stencil
        unique_indices = np.empty(0, dtype=np.int32)
        for velocities in schemes_velocities:
            unique_indices = np.union1d(unique_indices, velocities)
        self.unique_velocities = np.asarray(
            [Velocity(dim=self.dim, num=i) for i in unique_indices]
        )

        self.v = []
        self.nv = []
        self.nv_ptr = [0]
        num = self.unum
        for velocities in schemes_velocities:
            ypos = np.searchsorted(num, velocities)
            self.v.append(self.unique_velocities[ypos])
            size = len(velocities)
            self.nv.append(size)
            self.nv_ptr.append(self.nv_ptr[-1] + size)
        self.nv_ptr = np.asarray(self.nv_ptr)

        # get the index in the v[k] of the num velocity
        self.num2index = []
        for k in range(self.nstencils):
            num = self.num[k]
            self.num2index.extend(num)

        # get the index in the v[k] of the num velocity (unique)
        unum = self.unum
        self.unum2index = -1000 + np.zeros(np.max(unum) + 1, dtype=np.int32)
        self.unum2index[unum] = np.arange(unum.size)

        for k in range(self.nstencils):
            self.append(OneStencil(self.v[k], self.nv[k]))

        log.debug(self.__str__())

        # check if all the schemes are symmetric
        self.is_symmetric()

    @staticmethod
    def extract_dim(dico):
        """Extract the dimension from the dictionary"""
        dim = dico.get('dim', None)

        if not dim:
            dim, _ = get_box(dico)

        return dim

    @property
    def unvtot(self):
        """the number of unique velocities involved in the stencils."""
        return self.unique_velocities.size

    @property
    def vmax(self):
        """the maximal velocity in norm for each spatial direction."""
        tmp = np.asarray([self.uvx, self.uvy, self.uvz])
        return np.max(tmp[:self.dim], axis=1)

    @property
    def vmin(self):
        """the minimal velocity in norm for each spatial direction."""
        tmp = np.asarray([self.uvx, self.uvy, self.uvz])
        return np.min(tmp[:self.dim], axis=1)

    @property
    def vmax_full(self):
        """
        the maximal velocity in norm for each spatial direction.
        all the three dimensions are considered
        even if dim = 1 or 2
        """
        tmp = np.array([0, 0, 0])
        tmp[:self.dim] = self.vmax
        return tmp

    @property
    def uvx(self):
        """the x component of the unique velocities."""
        vectorize = np.vectorize(lambda obj: obj.vx)
        return vectorize(self.unique_velocities)

    @itemproperty
    def vx(self, k):
        """vx[k] the x component of the velocities for the stencil k."""
        vectorize = np.vectorize(lambda obj: obj.vx)
        return vectorize(self[k])

    @property
    def uvy(self):
        """the y component of the unique velocities."""
        vectorize = np.vectorize(lambda obj: obj.vy)
        return vectorize(self.unique_velocities)

    @itemproperty
    def vy(self, k):
        """vy[k] the y component of the velocities for the stencil k."""
        vectorize = np.vectorize(lambda obj: obj.vy)
        return vectorize(self[k])

    @property
    def uvz(self):
        """the z component of the unique velocities."""
        vectorize = np.vectorize(lambda obj: obj.vz)
        return vectorize(self.unique_velocities)

    @itemproperty
    def vz(self, k):
        """vz[k] the z component of the velocities for the stencil k."""
        vectorize = np.vectorize(lambda obj: obj.vz)
        return vectorize(self[k])

    @property
    def unum(self):
        """the numbering of the unique velocities."""
        vectorize = np.vectorize(lambda obj: obj.num)
        return vectorize(self.unique_velocities)

    @itemproperty
    def num(self, k):
        """num[k] the numbering of the velocities for the stencil k.
        """
        vectorize = np.vectorize(lambda obj: obj.num)
        return vectorize(self.v[k])

    def get_all_velocities(self, scheme_id=None):
        """
        get all the velocities for all the stencils in one array

        Parameters
        ----------
        scheme_id: int
            specify for which scheme we want all velocities
            if None, return the velocities for all the schemes

        Returns
        -------

        ndarray
            all velocities of a scheme or of all the schemes

        """
        if scheme_id is None:
            size = self.nv_ptr[-1]
            all_velocities = np.empty((size, self.dim), dtype='int')
            for vind in range(len(self)):
                vx = self.vx[vind]
                vy = self.vy[vind]
                vz = self.vz[vind]
                all_velocities[
                    self.nv_ptr[vind]:self.nv_ptr[vind+1], :
                ] = np.asarray([vx, vy, vz][:self.dim]).T
        else:
            vx = self.vx[scheme_id]
            vy = self.vy[scheme_id]
            vz = self.vz[scheme_id]
            all_velocities = np.asarray([vx, vy, vz][:self.dim]).T
        return all_velocities

    def get_symmetric(self, axis=None):
        """
        get the symmetric velocities.
        """
        ksym = np.empty(self.nv_ptr[-1], dtype=np.int32)

        k = 0
        for v in self.v:
            for vk in v:
                num = vk.get_symmetric(axis).num
                n = np.searchsorted(self.nv_ptr, k, side='right') - 1
                index = self.num2index[
                    self.nv_ptr[n]:self.nv_ptr[n+1]
                ].index(num) + self.nv_ptr[n]
                ksym[k] = index
                k += 1

        return ksym

    def __str__(self):
        from .utils import header_string
        from .jinja_env import env
        template = env.get_template('stencil.tpl')
        return template.render(
            header=header_string('Stencil information'),
            stencil=self
        )

    def __repr__(self):
        return self.__str__()

    def is_symmetric(self):
        """
        check if all the velocities have their symmetric.
        """
        for n in range(self.nstencils):
            is_sym = True
            scheme_velocities = np.array(self.num[n])
            for k in range(self.nv[n]):
                v = self.v[n][k]
                contains_sym = np.all(np.where(
                    scheme_velocities == v.get_symmetric().num, False, True
                ))
                if contains_sym:
                    err_msg = "The velocity {0} ".format(v)
                    err_msg += "has no symmetric velocity "
                    err_msg += "in the stencil {0:d}".format(n)
                    log.warning(err_msg)
                    is_sym = False
            if is_sym:
                log.info("The stencil %d is symmetric", n)
            else:
                log.warning("The stencil %d is not symmetric", n)
        return is_sym

    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    # pylint: disable=too-complex
    def visualize(self,
                  viewer_mod=viewer.matplotlib_viewer,
                  k=None,
                  unique_velocities=False,
                  view_label=True):
        """
        plot the velocities

        Parameters
        ----------

        viewer : package used to plot the figure (could be matplotlib, ...)
            see viewer for more information
        k : list of stencil index to plot
            if None plot all stencils
        unique_velocities : if True plot the unique velocities
        view_label: if True show the velocity numbering

        Returns
        -------

        fig
            the figure (fig if matplotlib is used)

        """
        def populate(vx, vy, vz):
            dummy = np.asarray([vx, vy, vz][:self.dim])
            if self.dim == 1:
                pos = np.zeros((dummy.size, 2))
                pos[:, 0] = dummy
            else:
                pos = dummy.T
            return pos

        pos = []
        title = []
        if unique_velocities:
            pos.append(populate(self.uvx, self.uvy, self.uvz))
            title.append("Stencil of the unique velocities")
        else:
            if k is None:
                schemes2plot = list(range(self.nstencils))
            elif isinstance(k, int):
                schemes2plot = [k]
            else:
                schemes2plot = k
            for i in schemes2plot:
                pos.append(populate(self.vx[i], self.vy[i], self.vz[i]))
                title.append("Stencil {0:d}".format(i))

        views = viewer_mod.Fig(
            len(pos), 1, dim=self.dim, figsize=(5, 5*len(pos))
        )
        views.fix_space(wspace=0.25, hspace=0.25)

        for i, posi in enumerate(pos):
            view = views[i]

            pminmax = np.zeros(6 if self.dim == 3 else 4)
            pminmax[::2] = np.min(posi, axis=0) - 1
            pminmax[1::2] = np.max(posi, axis=0) + 1
            view.axis(*pminmax, dim=self.dim, aspect='equal')

            view.title = title[i]
            view.xaxis(np.arange(pminmax[0], pminmax[1] + 1))
            if self.dim == 1:
                view.yaxis_set_visible(False)
            view.plot(pminmax[:2], [0, 0], color='orange', alpha=0.25)

            if self.dim >= 2:
                view.yaxis(np.arange(pminmax[2], pminmax[3] + 1))
                view.plot([0, 0], pminmax[2:4], color='orange', alpha=0.25)
            if self.dim == 3:
                view.zaxis(np.arange(pminmax[4], pminmax[5] + 1))
                view.plot(
                    [0, 0], [0, 0],
                    pminmax[4:], color='orange', alpha=0.25
                )

            if view_label:
                view.text(list(map(str, self.num[i])), posi,
                          fontsize=12, color='navy', fontweight='bold')
            view.grid(visible=True, which='major', alpha=0.25)

        views.show()
        return views



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Base element
"""

# pylint: disable=invalid-name

import logging
from abc import ABC, abstractmethod
from six.moves import range
import numpy as np

from .utils import distance_lines, distance_ellipse

__all__ = ['Element', 'BaseCircle', 'BaseEllipse',
           'BaseTriangle', 'BaseParallelogram']

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Element(ABC):
    """
    Class Element

    generic class for the elements

    """
    number_of_bounds = -1

    def __init__(self, label, isfluid):
        self.isfluid = isfluid
        if isinstance(label, int):
            self.label = [label]*self.number_of_bounds
        else:
            self.label = label
        self.test_label()

    @abstractmethod
    def get_bounds(self):
        """
        return the smallest box where the element is.

        """

    @abstractmethod
    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the element.

        Notes
        -----

        the edges of the element are considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the element, False otherwise)

        """

    @abstractmethod
    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(2), alpha=1.
                  ):
        """
        visualize the element

        Parameters
        ----------

        viewer : Viewer
            a viewer (default matplotlib_viewer)

        color : color
            color of the element

        viewlabel : bool
            activate the labels mark (default False)

        scale : ndarray
            scale the distance of the labels (default ones)

        alpha : double
            transparency of the element (default 1)

        """

    def __repr__(self):
        return self.__str__()

    def test_label(self):
        """
        test if the number of labels is equal to the number of bounds.
        """
        return len(self.label) == self.number_of_bounds


class BaseCircle:
    """
    Class BaseCircle

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the three coordinates of the first vector defining the circular base
    v2 : list
        the three coordinates of the second vector defining the circular base

    """
    def __init__(self, center, v1, v2):
        self.center = np.asarray(center)
        radius = np.linalg.norm(v1)
        if radius != np.linalg.norm(v2):
            err_msg = "Error in BaseCircle: "
            err_msg += "vectors v1 and v2 must have the same norm"
            log.error(err_msg)
        self.radius = radius
        # orthogonalization of the two vectors
        self.v1 = np.asarray(v1)
        v2 = np.asarray(v2)
        self.v2 = v2 - np.inner(v2, self.v1) * self.v1 \
            / np.inner(self.v1, self.v1)
        nv2 = np.linalg.norm(self.v2)
        if nv2 == 0:
            err_msg = "Error in the definition of the cylinder: "
            err_msg += "the vectors are colinear"
            log.error(err_msg)
        log.info(self.__str__())

    def get_bounds(self):
        """
        Get the bounds of the base
        """
        return self.center - self.radius, self.center + self.radius

    # pylint: disable=no-self-use
    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the element.

        Notes
        -----

        the edges of the element are considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the element, False otherwise)

        """
        x, y = grid
        return (x**2 + y**2) <= 1.

    @staticmethod
    def distance(grid, v, dmax, label):
        """
        Compute the distance in the v direction between the element
        and the points defined by (x, y) for a given label.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max
        label : int
            the label of interest

        Returns
        -------

        ndarray
            array of distances

        """
        x, y = grid
        c = np.zeros((2,))
        v1 = np.asarray([1, 0])
        v2 = np.asarray([0, 1])
        return distance_ellipse(x, y, v, c, v1, v2, dmax, label)

    @staticmethod
    def _visualize():
        t = np.linspace(0, 2.*np.pi, 100)
        lx_b = [np.cos(t), np.cos(t), np.cos(t[::-1])]
        ly_b = [np.sin(t), np.sin(t), np.sin(t[::-1])]
        return lx_b, ly_b

    def __str__(self):
        s = "Circular base with radius {} ".format(self.radius)
        s += "centered in " + str(self.center) + "\n"
        s += "     in the plane spanned by " + str(self.v1)
        s += " and " + str(self.v2) + "\n"
        return s


class BaseEllipse:
    """
    Class BaseEllipse

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the three coordinates of the first vector defining the ellipsoidal base
    v2 : list
        the three coordinates of the second vector
        defining the ellipsoidal base

    Warnings
    --------

    The vectors v1 and v2 have to be orthogonal.

    """

    def __init__(self, center, v1, v2):
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        if abs(np.inner(self.v1, self.v2)) > 1.e-10:
            err_msg = "Error in the definition of the cylinder: "
            err_msg += "the vectors have to be orthogonal"
            log.error(err_msg)
        log.info(self.__str__())

    def get_bounds(self):
        """
        Get the bounds of the base
        """
        r = max(np.linalg.norm(self.v1), np.linalg.norm(self.v2))
        return self.center - r, self.center + r

    # pylint: disable=no-self-use
    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the element.

        Notes
        -----

        the edges of the element are considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the element, False otherwise)

        """
        x, y = grid
        return (x**2 + y**2) <= 1.

    @staticmethod
    def distance(grid, v, dmax, label):
        """
        Compute the distance in the v direction between the element
        and the points defined by (x, y) for a given label.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max
        label : int
            the label of interest

        Returns
        -------

        ndarray
            array of distances

        """
        x, y = grid
        c = np.zeros((2,))
        v1 = np.asarray([1, 0])
        v2 = np.asarray([0, 1])
        return distance_ellipse(x, y, v, c, v1, v2, dmax, label)

    @staticmethod
    def _visualize():
        t = np.linspace(0, 2.*np.pi, 100)
        lx_b = [np.cos(t), np.cos(t), np.cos(t[::-1])]
        ly_b = [np.sin(t), np.sin(t), np.sin(t[::-1])]
        return lx_b, ly_b

    def __str__(self):
        s = 'Ellipsoidal base centered in ' + str(self.center) + '\n'
        s += '     in the plane spanned by ' + str(self.v1)
        s += ' and ' + str(self.v2) + '\n'
        return s


class BaseTriangle:
    """
    Class BaseTriangle

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the three coordinates of the first vector defining the triangular base
    v2 : list
        the three coordinates of the second vector defining the triangular base

    """

    def __init__(self, center, v1, v2):
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        nv1 = np.linalg.norm(self.v1)
        nv2 = np.linalg.norm(self.v2)
        if np.allclose(nv1*self.v2, nv2*self.v1):
            err_msg = "Error in the definition of the cylinder: "
            err_msg += "the vectors are not free"
            log.error(err_msg)
        log.info(self.__str__())

    def get_bounds(self):
        """
        Get the bounds of the base
        """
        box = np.asarray(
            [
                self.center,
                self.center + self.v1,
                self.center + self.v1 + self.v2,
                self.center + self.v2
            ]
        )
        return np.min(box, axis=0), np.max(box, axis=0)

    # pylint: disable=no-self-use
    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the element.

        Notes
        -----

        the edges of the element are considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the element, False otherwise)

        """
        x, y = grid
        return np.logical_and(np.logical_and(x >= 0, y >= 0), x + y <= 1)

    @staticmethod
    def distance(grid, v, dmax, label):
        """
        Compute the distance in the v direction between the element
        and the points defined by (x, y) for a given label.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max
        label : int
            the label of interest

        Returns
        -------

        ndarray
            array of distances

        """
        x, y = grid
        p = [[0, 0], [0, 0], [1, 0]]
        vt = [[1, 0], [0, 1], [-1, 1]]
        return distance_lines(x, y, v, p, vt, dmax, label)

    @staticmethod
    def _visualize():
        p = np.asarray([[0, 0],
                        [1, 0],
                        [0, 1],
                        [0, 0],
                        [1, 0]]
                       ).T
        lx_b = []
        ly_b = []
        for k in range(3):
            lx_b.append(p[0, k:k+2])
            ly_b.append(p[1, k:k+2])
        lx_b.append(p[0, :4])
        ly_b.append(p[1, :4])
        lx_b.append(p[0, 3::-1])
        ly_b.append(p[1, 3::-1])
        return lx_b, ly_b

    def __str__(self):
        s = 'Triangular base centered in ' + str(self.center) + '\n'
        s += '     in the plane spanned by ' + str(self.v1)
        s += ' and ' + str(self.v2) + '\n'
        return s


class BaseParallelogram:
    """
    Class BaseParallelogram

    Parameters
    ----------
    center : list
        the three coordinates of the center
    v1 : list
        the three coordinates of the first vector that defines the base
    v2 : list
        the three coordinates of the second vector that defines the base

    """
    def __init__(self, center, v1, v2):
        self.center = np.asarray(center)
        self.v1 = np.asarray(v1)
        self.v2 = np.asarray(v2)
        nv1 = np.linalg.norm(self.v1)
        nv2 = np.linalg.norm(self.v2)
        if np.allclose(nv1*self.v2, nv2*self.v1):
            err_msg = "Error in the definition of the cylinder: "
            err_msg += "the vectors are not free"
            log.error(err_msg)
        log.info(self.__str__())

    def get_bounds(self):
        """
        Get the bounds of the base
        """
        box = np.asarray(
            [
                self.center,
                self.center + self.v1,
                self.center + self.v1 + self.v2,
                self.center + self.v2
            ]
        )
        return np.min(box, axis=0), np.max(box, axis=0)

    # pylint: disable=no-self-use
    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the element.

        Notes
        -----

        the edges of the element are considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the element, False otherwise)

        """
        x, y = grid
        return np.logical_and(np.logical_and(x >= 0, y >= 0),
                              np.logical_and(x <= 1, y <= 1))

    @staticmethod
    def distance(grid, v, dmax, label):
        """
        Compute the distance in the v direction between the element
        and the points defined by (x, y) for a given label.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max
        label : int
            the label of interest

        Returns
        -------

        ndarray
            array of distances

        """
        x, y = grid
        p = [[0, 0], [0, 0], [1, 0], [0, 1]]
        vt = [[1, 0], [0, 1], [0, 1], [1, 0]]
        return distance_lines(x, y, v, p, vt, dmax, label)

    @staticmethod
    def _visualize():
        p = np.asarray([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0], [1, 0]]).T
        lx_b = []
        ly_b = []
        for k in range(4):
            lx_b.append(p[0, k:k+2])
            ly_b.append(p[1, k:k+2])
        lx_b.append(p[0, :5])
        ly_b.append(p[1, :5])
        lx_b.append(p[0, 4::-1])
        ly_b.append(p[1, 4::-1])
        return lx_b, ly_b

    def __str__(self):
        s = 'Parallelogram base centered in ' + str(self.center) + '\n'
        s += '     in the plane spanned by ' + str(self.v1)
        s += ' and ' + str(self.v2) + '\n'
        return s



# FIXME: make pylint happy !
# pylint: disable=all

import collections
from .codegen import make_routine
from .autowrap import autowrap


class Generator:
    def __init__(self, backend, directory=None, verbose=False):
        self.routines = collections.OrderedDict()
        self.module = None
        self.directory = directory
        self.backend = backend
        self.verbose = verbose

    def add_routine(self, name_expr,
                    local_vars=None, settings={}):
        self.routines[name_expr[0]] = make_routine(name_expr[0], name_expr[1],
                                                   user_local_vars=local_vars,
                                                   language=self.backend,
                                                   settings=settings)

    def compile(self):
        self.module = autowrap(self.routines.values(),
                               self.backend,
                               self.directory,
                               verbose=self.verbose)
